from .util import box_transform, box_transform_inv, clip_boxes
import torch.nn.functional as F
import numpy as np
import torch
from torch.autograd import Variable
try:
    from lung_analysis_pipeline.NoduleDetect_SANet.utils.pybox import *
except ImportError:
    print('Warning: C++ module import failed! This should only happen in deployment')
    from lung_analysis_pipeline.NoduleDetect_SANet.utils.util import py_nms as torch_nms
    from lung_analysis_pipeline.NoduleDetect_SANet.utils.util import py_box_overlap as torch_overlap


def rcnn_encode(window, truth_box, weight):
    return box_transform(window, truth_box, weight)


def rcnn_decode(window, delta, weight):
    return  box_transform_inv(window, delta, weight)


def rcnn_nms(cfg, mode, inputs, proposals, logits, deltas):

    if mode in ['train',]:
        nms_pre_score_threshold = cfg['rcnn_train_nms_pre_score_threshold']
        nms_overlap_threshold   = cfg['rcnn_train_nms_overlap_threshold']

    elif mode in ['valid', 'test','eval']:
        nms_pre_score_threshold = cfg['rcnn_test_nms_pre_score_threshold']
        nms_overlap_threshold   = cfg['rcnn_test_nms_overlap_threshold']
        #
        # if mode in ['eval']:
        #     nms_pre_score_threshold = 0.05 # set low numbe r to make roc curve.

    else:
        raise ValueError('rcnn_nms(): invalid mode = %s?'%mode)


    batch_size, _, depth, height, width = inputs.size() #original image width
    num_class = cfg['num_class']

    # probs     = np_sigmoid(logits.cpu().data.numpy())
    # probs = F.softmax(logits).cpu().data.numpy()
    probs = F.softmax(logits, dim=1).cpu().data.numpy()

    deltas = deltas.cpu().data.numpy().reshape(-1, num_class, 6)
    proposals = proposals.cpu().data.numpy()
    # masks = (F.sigmoid(mask_logits).cpu().data.numpy() > 0.5).astype(np.uint8)

    #non-max suppression
    detections = []
    # segments = []
    keeps = []
    for b in range(batch_size):
        detection = [np.empty((0, 9), np.float32),]

        index = np.where(proposals[:,0] == b)[0]
        if len(index)>0:
            prob  = probs[index]
            delta = deltas[index]
            proposal = proposals[index]
            # mask = masks[index]
            # cats = np.argmax(prob, 1)

            for j in range(1, num_class): #skip background
                idx = np.where(prob[:, j] > nms_pre_score_threshold)[0]
                # idx = np.where(cats == j)[0]
                if len(idx)>0:
                    p = prob[idx, j].reshape(-1, 1)
                    d = delta[idx, j]
                    # m = mask[idx, j - 1]
                    box = rcnn_decode(proposal[idx, 2:8], d, cfg['box_reg_weight'])
                    box = clip_boxes(box, inputs.shape[2:])
                    # box = clip_boxes(box, width, height)

                    # keep = filter_boxes(box, min_size = nms_min_size)
                    # num  = len(keep)
                    # if num>0:
                        # box  = box[keep]
                        # p    = p[keep]
                    js = np.expand_dims(np.array([j] * len(p)), axis=-1)
                    output = np.concatenate((p, box, js), 1)

                    if len(output) > 0:
                        output = torch.from_numpy(output).float()
                        output, keep = torch_nms(output, nms_overlap_threshold)

                    num = len(output)

                    if num > 0:
                        det = np.zeros((num, 9),np.float32)
                        det[:, 0] = b
                        det[:, 1:] = output
                        detection.append(det)
                        # segments.append(m[keep.numpy()])
                        keeps.extend(index[idx[keep.numpy()]].tolist())

        detection = np.vstack(detection)

        detections.append(detection)

    detections = Variable(torch.from_numpy(np.vstack(detections))).cuda()
    # segments = np.vstack(segments)
    return detections, keeps


def get_probability(cfg, mode, inputs, proposals, logits, deltas):
    if mode in ['train',]:
        nms_pre_score_threshold = cfg['rcnn_train_nms_pre_score_threshold']
        nms_overlap_threshold   = cfg['rcnn_train_nms_overlap_threshold']

    elif mode in ['valid', 'test','eval']:
        nms_pre_score_threshold = cfg['rcnn_test_nms_pre_score_threshold']
        nms_overlap_threshold   = cfg['rcnn_test_nms_overlap_threshold']
    else:
        raise ValueError('rcnn_nms(): invalid mode = %s?'%mode)

    num_class = cfg['num_class']
    # probs = F.softmax(logits).cpu().data.numpy()
    probs = F.softmax(logits, dim=1).cpu().data.numpy()
    deltas = deltas.cpu().data.numpy().reshape(-1, num_class, 6)
    proposals = proposals.cpu().data.numpy()
    
    output = None 
    for j in range(1, num_class):  # skip background
        idx = np.where(probs[:, j] > nms_pre_score_threshold)[0]
        if len(idx) > 0:
            p = probs[idx, j].reshape(-1, 1)
            d = deltas[idx, j]
            box = rcnn_decode(proposals[idx, 2:8], d, cfg['box_reg_weight'])
            box = clip_boxes(box, inputs.shape[2:])
            js = np.expand_dims(np.array([j] * len(p)), axis=-1)
            output = np.concatenate((p, box, js), 1)
    if output is None: 
        output = np.empty((0, 0), dtype=np.float32)

    return torch.from_numpy(output).cuda().float()


#-----------------------------------------------------------------------------  
if __name__ == '__main__':
    print( '%s: calling main function ... ' % os.path.basename(__file__))



 



import torch
import torch.nn as nn


# Define GAN loss: [gan | lsgan | wgan-gp | wgan-gp0 | hinge ]
class GANLoss(nn.Module):
    def __init__(self, gan_type, real_label_val=1.0, fake_label_val=0.0):
        super(GANLoss, self).__init__()
        self.gan_type = gan_type.lower()
        self.real_label_val = real_label_val
        self.fake_label_val = fake_label_val

        print("gan type is:", self.gan_type)
        print("real label value:", self.real_label_val)
        print("fake lable value:", self.fake_label_val)

        if self.gan_type == 'gan' or self.gan_type == 'ragan':
            self.loss = nn.BCEWithLogitsLoss()
        elif self.gan_type == 'lsgan':
            self.loss = nn.MSELoss()
        elif 'wgan' in self.gan_type:
            def _wgan_loss(input, target):
                # target is boolean
                return -1 * input.mean() if target else input.mean()
            self.loss = _wgan_loss

        # if 'gan_type' is 'hinge', we define the '_hing_loss' function and store the function as an attribute
        elif self.gan_type =='hinge':
            def _hinge_loss(input, target):
                # print("hinge loss caled")
                # print("target:", target)
                if target:
                    return nn.ReLU()(1.0 - input).mean()
                else: 
                    return nn.ReLU()(1.0 + input).mean()
                # return -1 * input.mean() if target else input.mean()
            self.loss = _hinge_loss
        else:
            raise NotImplementedError('GAN type [{:s}] is not found'.format(self.gan_type))

    def get_target_label(self, input, target_is_real):
        # print("get_target_label is:", target_is_real)
        # if 'self.gan_type' is 'wgan' or 'hinge', we do not need target label 
        if 'wgan' in self.gan_type or self.gan_type == 'hinge':
            # print("{} returned".format(target_is_real))
            return target_is_real
        # otherwise, return tensor filled with 1 if real label or 0 if fake label
        if target_is_real:
            return torch.empty_like(input).fill_(self.real_label_val)
        else:
            return torch.empty_like(input).fill_(self.fake_label_val)

    def forward(self, input, target_is_real):
        target_label = self.get_target_label(input, target_is_real)
        loss = self.loss(input, target_label)
        return loss


# gradient penalty regularization
# wgan-gp or zero centered penalty https://github.com/LMescheder/GAN_stability
class GradientPenaltyLoss(nn.Module):
    def __init__(self, center):
        super(GradientPenaltyLoss, self).__init__()
        self.center = center

    @staticmethod
    def compute_grad2(x_in, d_out):
        batch_size = x_in.size(0)
        grad_dout = torch.autograd.grad(
            outputs=d_out.sum(), inputs=x_in,
            create_graph=True, retain_graph=True, only_inputs=True)[0]
        grad_dout2 = grad_dout.pow(2)
        assert (grad_dout2.size() == x_in.size())
        reg = grad_dout2.view(batch_size, -1).sum(dim=1)
        return reg

    def gp_reg(self, x_in, d_out):
        if self.center == 0.:
            return self.compute_grad2(x_in, d_out).mean()
        else:
            return (self.compute_grad2(x_in, d_out).sqrt() - self.center).pow(2).mean()

    def forward(self, x_in, d_out):
        return self.gp_reg(x_in, d_out)



import os, time, glob
import math
from datetime import datetime
import pydicom
import numpy as np
import cv2
import random
import torch
import logging
import nrrd
import json
import csv
import SimpleITK as sitk
import re
import nibabel as nib
import torch.nn.functional as F
from scipy import ndimage


"""
Get current timestamp in ymd-HMS format. Used when creataing a directory that already exists
"""
def get_timestamp():
    return datetime.now().strftime('%y%m%d-%H%M%S')

"""
Create a folder if it doesn't exist
"""
def mkdir(path):
    if not os.path.exists(path):
        os.makedirs(path)

"""
Make mupltiple directories, if it is an instance of string
"""
def mkdirs(paths):
    if isinstance(paths, str):
        mkdir(paths)
    else:
        for path in paths:
            mkdir(path)

"""
Create a folder if it doesn't exits, otherwise create a renamed version if it already exists
"""
def mkdir_and_rename(path):
    if os.path.exists(path):
        new_name = path + '_archived_' + get_timestamp()
        print('Path already exists. Rename it to [{:s}]'.format(new_name))
        logger = logging.getLogger('base')
        logger.info('Path already exists. Rename it to [{:s}]'.format(new_name))
        os.rename(path, new_name)
    os.makedirs(path)


"""
Set seed for numpy and torch arrays
"""
def set_random_seed(seed):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)


"""
Create two logger named 'base' and 'val', with file name specified in 'phase' variable at the 
path specified in 'root' variable.
If 'screen=True', message is printed to screen
"""
def setup_logger(logger_name, root, phase, level=logging.INFO, screen=False):
    '''set up logger'''
    l = logging.getLogger(logger_name)
    formatter = logging.Formatter(
        '%(asctime)s.%(msecs)03d - %(levelname)s: %(message)s', datefmt='%y-%m-%d %H:%M:%S')
    log_file = os.path.join(root, phase + '_{}.log'.format(get_timestamp()))
    fh = logging.FileHandler(log_file, mode='w')
    fh.setFormatter(formatter)
    l.setLevel(level)
    l.addHandler(fh)
    if screen:
        sh = logging.StreamHandler()
        sh.setFormatter(formatter)
        l.addHandler(sh)

"""
Prints evaluaation metrics on screen
"""
def print_metrics(logger, metric_name, results):
    means = {}
    for view in list(results[list(results.keys())[0]].keys()):
        count = 0
        sum = 0.
        for k, v in results.items():
            count += 1
            sum += v[view]
            logger.info('{} for {} in {} view : {:.6f}'.format(metric_name, k, view, v[view]))
        means[view] = sum/count
        logger.info('Average {} in {} view : {:.6f}'.format(metric_name, view, means[view]))
    return means


####################
# image operations
####################
def tensor2img(tensor, out_type=np.uint8, min_max=(0, 1), intercept=-1024):
    '''
    Converts a torch Tensor into an image Numpy array
    Input: 3D(D,H,W)
    Output: 3D(H,W,D), np.uint8 (default), [0,1500] uint16, [-1000, 500] int16
    intercept is used when out_type is set to int16
    '''
    tensor = tensor.cpu().float().clamp_(*min_max)  # clamp
    tensor = (tensor - min_max[0]) / (min_max[1] - min_max[0])  # to range [0,1]
    img_np = tensor.numpy()
    if out_type == np.uint8:
        img_np = (img_np * 255.0).round()
        # Important. Unlike matlab, numpy.uint8() WILL NOT round by default.
    if out_type == np.uint16:
        img_np = (img_np * 1500.0).round()
    if out_type == np.int16:
        img_np = (img_np * 1500.0).round() + intercept 
    return img_np.astype(out_type)

def tensor2mask(tensor, out_type=np.uint8, min_max=(0, 1), intercept=-1024):
    '''
    Converts a torch Tensor into a mask Numpy array
    Input: 3D(D,H,W)
    Output: 3D(H,W,D), np.uint8 (default), [0,1] uint16
    '''
    mask_np = tensor.cpu().numpy()
    return mask_np.astype(out_type)


def img2tensor(image, cent=1., factor=255./2.):
    '''
    :Input 3D(H,W,D):
    :Output (1,D,H,W) [-1,1]
    '''
    return torch.Tensor((image / factor - cent)[:, :, :, np.newaxis].transpose((3, 2, 0, 1)))


"""
Converts incoming h5 data input (uint16 - D,H,W) to torch tensor float in a given range range [0, 1500].
For dataloader use, inplace operation
"""
class ImgToTensor(object):
    def __call__(self, sample, clip=False, raw_data_range=1500.):
        if clip:
            img = torch.from_numpy(np.clip(np.array(sample, np.float32, copy=False), 0, raw_data_range))
        else:
            img = torch.from_numpy(np.array(sample, np.float32, copy=False))
        return img.float().div_(raw_data_range)  # range is from 0 to 1500 for png


"""
Read json configuration file for each image and returns the spacing, orientation and origin of the image
"""
def read_config(config_path):
    config_path = os.path.join(config_path) 
    with open(config_path, 'r') as f:
        config = json.load(f)
    meta_data = {}
    meta_data['Spacing'] = [float(i) for i in config['Spacing'].split()]
    meta_data['Orientation'] = [float(i) for i in config['Orientation'].split()] 
    meta_data['Origin'] = [float(i) for i in config['Origin'].split()]
    return meta_data

def save_vol(opt, spacings, results_dir, patient_id, volume, dtype):
    # nrrd accepts in [W,H,D] format
    # [D,H,W]==>[W,H,D]
    # no need to transpose, just add index_order='C'
    # print(volume.shape)
    # volume = np.transpose(volume, (2, 1, 0))
    # print(volume.shape)
    if dtype == 'nrrd': 
        if len(spacings) != 0:
            copy_spacings = spacings.copy()
            copy_spacings[2] /= opt['scale']
        else:
            copy_spacings = []

        volume_path = os.path.join(results_dir, patient_id+'.nrrd')
        nrrd.write(volume_path, volume, index_order='C')
    elif dtype == 'nifti': 
        volume_path = os.path.join(results_dir, patient_id+'.nii.gz')
        volume = np.transpose(volume, (2, 1, 0))
        volume_img = nib.Nifti1Image(volume, np.eye(4))
        # Save the image to a file
        nib.save(volume_img, volume_path)
    else: 
        raise NotImplementedError('Output dtype not supported.')

def save_mask(opt, input_info, results_dir, seg_result_dtype, patient_id, mask):
    
    mask = mask.astype(np.uint8) # to unsigned char 
    if seg_result_dtype == 'nrrd': 
        volume_path = os.path.join(results_dir, patient_id+'.nrrd')

        header = {
            'units': ['mm', 'mm', 'mm'], 
            'Segment0_LabelValue': 1, 
        } 

        # save header in dict format 
        # spacings = input_info['spacings']
        # space_directions = ((spacings[0], 0, 0), (0, spacings[1], 0), (0, 0, spacings[2]))
        # header['space directions'] = space_directions
        # header['space origin'] = input_info['image_position']

        nrrd.write(volume_path, mask, header, index_order='C')
    elif seg_result_dtype == 'nifti': 
        volume_path = os.path.join(results_dir, patient_id+'.nii.gz')
        # print('mask shape before: {}'.format(mask.shape))
        mask = np.transpose(mask, (2, 1, 0))
        # print('mask shape after: {}'.format(mask.shape))
        mask_img = nib.Nifti1Image(mask, np.eye(4))

        # Save the image to a file
        nib.save(mask_img, volume_path)
    else: 
        raise NotImplementedError('Segmentation output dtype not supported.')

def save_vol_nested(opt, spacings, results_dir, input_info, volume):
    patient_id = input_info['patient_id'][0]
    time = input_info['time'][0]
    name = input_info['name'][0]
    # create nested directories if not exist 
    patient_dir = os.path.join(results_dir, patient_id)
    mkdir(patient_dir)
    time_dir = os.path.join(patient_dir, time)
    mkdir(time_dir)
    name_dir = os.path.join(time_dir, name)
    mkdir(name_dir) 
    mkdir(os.path.join(name_dir, 'NRRD'))
    volume_path = os.path.join(name_dir, 'NRRD', '1.nrrd')
    # nrrd accepts in [W,H,D] format
    # [D,H,W]==>[W,H,D]
    # no need to transpose, just add index_order='C'
    # volume = np.transpose(volume, (2, 1, 0))
    if len(spacings) != 0:
        copy_spacings = spacings.copy()
        copy_spacings[2] /= opt['scale']
    else:
        copy_spacings = []

    header = {'units': ['mm', 'mm', 'mm'], 'spacings': copy_spacings}
    nrrd.write(volume_path, volume, header, index_order='C')


def save_dicoms(opt, spacings, volume_path, volume):
    mkdir(volume_path)
    copy_spacings = spacings.copy()
    copy_spacings[2] /= opt['scale'] 
    # if opt['datasets']['val']['data_type'] == 'h5':     
    #     config_path =  os.path.join(opt['datasets']['val']['dataroot_LR'], uid + '.json')
    #     meta_data = read_config(config_path)
    #     spacings = meta_data['Spacing']
    #     spacings[2] /= opt['scale'] 
    #     orientations = meta_data['Orientation']
    #     orientations = [orientations[i:i+3] for i in range(0, len(orientations), 3)]
    #     origin = meta_data['Origin']  
    # elif opt['datasets']['val']['data_type'] == 'dicom':
    #     spacings[2] /= opt['scale'] 
    # else: 
    #     raise NotImplementedError('supported output format: nrrd or dicom')
    write_dicom(copy_spacings, volume, volume_path)            
    print("written as dcm completed")
    # for i, pixel_array in enumerate(vol): 
    #     write_dicom(i+1, opt, meta_data, pixel_array, volume_path)

def save_features(feature_type, opt, data): 
    feat_dtype = opt['feature_extraction'][feature_type]['output_dtype']
    feat_file = opt['feature_extraction'][feature_type]['output_path']
    id = data['uid'][0]
    if feature_type == 'deep_feature':
        save_feature_csv(feat_file, feat_dtype, id, data['deep_feature'].cpu().numpy())
    elif feature_type == 'radiomics':
        save_feature_json(feat_file, feat_dtype, id, data['radiomics']) 
    else: 
        raise NotImplementedError('Feature extraction output dtype not supported.')

def save_feature_json(filename, feat_dtype, id, features): 
    # Check if the file exists and create it if it doesn't
    results_dir = os.path.dirname(filename)
    mkdir(results_dir)
    # Check if the JSON file exists
    if os.path.exists(filename):
        # File exists, load its contents
        with open(filename, 'r') as file:
            existing_data = json.load(file)
    else:
        # File doesn't exist, create an empty object
        existing_data = []
    
    existing_data.extend(features)

    # Write the updated data to the JSON file
    with open(filename, 'w') as file:
        json.dump(existing_data, file)

def save_feature_csv(feat_output_file, feat_dtype, id, features): 
    """
    param feat_output_file: path to the csv file that the features are saved to. 
    param feature_dtype: format of the output file. Only support csv for now. 
    param id: image id.
    param features: a numpy array of size [num_nodule, 256]. 
    This function appends features to the specified csv file.     
    """
    # Check if the file exists and create it if it doesn't
    results_dir = os.path.dirname(feat_output_file)
    mkdir(results_dir)
    if not os.path.exists(feat_output_file):
        with open(feat_output_file, 'w', newline='') as f:
            writer = csv.writer(f)
            header_row = ["ID"]
            for i in range(256):
                header_row.append(f"Feature_{i}")
            writer.writerow(header_row)
    with open(feat_output_file, 'a', newline='') as f:
        writer = csv.writer(f)
        for i in range(len(features)):
            feature_id = f"{id}-{i}"
            row = [feature_id]
            for j in range(256):
                row.append(features[i][j])
            writer.writerow(row)
    

def save_detect_bbox(opt, data): 
    """
    Saves the output of nodule detection model to a given file. Only csv supported. 
    """

    if opt['detection']['output_dtype'] == 'csv': 
        # output location 
        output_location = opt['detection']['output_path']
        results_dir = os.path.dirname(output_location)
        mkdir(results_dir)
        
        # data to be saved 
        if data is None: 
            print('data is None')
            return 
        if data['detection_bboxes'] is  None:
            return 
        
        detection_bboxes = data['detection_bboxes']
        
        # create the file if not exists 
        if not os.path.exists(output_location):
            with open(output_location, 'w', newline='') as f:
                writer = csv.writer(f)
                header_row = ['ID','center_z','center_y','center_x', 'diameter', 'probability']
                

                writer.writerow(header_row)
        
        # write data to file 
        with open(output_location, 'a', newline='') as f:
            writer = csv.writer(f)
            uid = data['uid'][0]
            for i in range(len(detection_bboxes)):
                # generate a bounding box id 
                bbox_id = f"{uid}-{i}"
                # [center_z, center_y, center_x]
                z, y, x, d, p = detection_bboxes[i][1], detection_bboxes[i][2], detection_bboxes[i][3], detection_bboxes[i][4], detection_bboxes[i][0]
                row = [bbox_id, z, y, x, d, p]
                # save the row 
                writer.writerow(row)
    else: 
        raise NotImplementedError('Nodule detection output dtype not supported.')

def write_dicom(spacings, new_arr, path):
    new_img = sitk.GetImageFromArray(new_arr)
    new_img.SetSpacing(spacings)
    
    writer = sitk.ImageFileWriter()
    # Use the study/series/frame of reference information given in the meta-data
    # dictionary and not the automatically generated information from the file IO
    writer.KeepOriginalImageUIDOn()

    modification_time = time.strftime("%H%M%S")
    modification_date = time.strftime("%Y%m%d")

    # Copy some of the tags and add the relevant tags indicating the change.
    # For the series instance UID (0020|000e), each of the components is a number, cannot start
    # with zero, and separated by a '.' We create a unique series ID using the date and time.
    # tags of interest:
    direction = new_img.GetDirection()
    series_tag_values = [("0008|0031",modification_time), # Series Time
                    ("0008|0021",modification_date), # Series Date
                    ("0008|0008","DERIVED\\SECONDARY"), # Image Type
                    ("0020|000e", "1.2.826.0.1.3680043.2.1125."+modification_date+".1"+modification_time), # Series Instance UID
                    ("0020|0037", '\\'.join(map(str, (direction[0], direction[3], direction[6],# Image Orientation (Patient)
                                                    direction[1],direction[4],direction[7])))),
                    ("0008|103e", "Normalized image"), # series description
                    ("0020|000d", "1.2.826.0.1.3680043.2.1125."+modification_date+".1"+modification_time), # study instance UID
                    ("0008|0020", modification_date), # study date
                    ("0008|0030", modification_time),  # study time
                    # Setting the type to CT preserves the slice location.
                    # set the type to CT so the thickness is carried over
                    ("0008|0060", "CT"),
                    ("0028|1050", "-600"), # WindowCenter
                    ("0028|1051", "1500"), # WindowWidth
                    ("0028|1054", "HU"), # RescaleType
                    ("0018|0050", str(spacings[2]))  # slice thickness
                    ]  

    # Write slices to output directory
    list(map(lambda i: writeSlices(writer, i, series_tag_values, new_img, path), range(new_img.GetDepth())))

def writeSlices(writer, index, series_tag_values, new_img, path):
    # DicomSeriesFromArray
    #https://simpleitk.readthedocs.io/en/next/Examples/DicomSeriesFromArray/Documentation.html
    image_slice = new_img[:,:,index]

    # Tags shared by the series.
    list(map(lambda tag_value: image_slice.SetMetaData(tag_value[0], tag_value[1]), series_tag_values))

    # Slice specific tags.
    image_slice.SetMetaData("0008|0012", time.strftime("%Y%m%d")) # Instance Creation Date
    image_slice.SetMetaData("0008|0013", time.strftime("%H%M%S")) # Instance Creation Time

    # (0020, 0032) image position patient determines the 3D spacing between slices.
    pos = [str(i) for i in new_img.TransformIndexToPhysicalPoint((0,0,index))]
    image_slice.SetMetaData("0020|0032", '\\'.join(pos)) # Image Position (Patient)
    image_slice.SetMetaData("0020|1041", pos[2]) # Slice location
    image_slice.SetMetaData("0020|0013", str(index+1)) # Instance Number

    # No need to set intercept/slope in ITK
    # RescaleIntercept 
    # image_slice.SetMetaData("0028|1052", "-1000")
    # RescaleSlope
    # image_slice.SetMetaData("0028|1053", "1")
    
    # Write to the output directory and add the extension dcm, to force writing in DICOM format.
    writer.SetFileName(os.path.join(path, '{0:0=3d}.dcm'.format(index+1)))
    writer.Execute(image_slice)

# need some work to get it working using pydicom
def write_pydicom(index, opt, meta_data, pixel_array, path):
    fullpath = os.path.join(path, '{0:0=3d}.dcm'.format(index))
    pixel_array = pixel_array.astype(np.int16)

    file_meta = Dataset()
    file_meta.MediaStorageSOPClassUID = '1.2.840.10008.5.1.4.1.1.2'
    file_meta.MediaStorageSOPInstanceUID = "1.2.840.113654.2.55.102869246940549636091903990697209280630"
    file_meta.ImplementationClassUID = '1.2.40.0.13.1.1.1'
    file_meta.TransferSyntaxUID = '1.2.840.10008.1.2.1' 

    ds = FileDataset(fullpath, {},file_meta = file_meta, preamble=b"\0"*128)

    # ds.SOPClassUID = '1.2.840.10008.5.1.4.1.1.2'
    # ds.SOPInstanceUID = "1.2.840.113654.2.55.102869246940549636091903990697209280630"
    ds.SeriesDescription = 'Normalized images'
    ds.SeriesInstanceUID = '1.2.840.113654.2.55.102869246940549636091903990697209280630'
    # ds.StudyInstanceUID =  '1.2.840.113654.2.55.102869246940549636091903990697209280630'
    # ds.StudyID = ''
    # ds.ContentDate = str(datetime.date.today()).replace('-','')
    # ds.ContentTime = str(time.time()) #milliseconds since the epoch
    ds.Modality = "CT"
    ds.SamplesPerPixel = 1
    ds.PhotometricInterpretation = "MONOCHROME2"
    ds.PixelRepresentation = 0
    ds.HighBit = 11
    ds.BitsStored = 12
    ds.BitsAllocated = 16
    ds.Columns, ds.Rows = pixel_array.shape
    ds.SeriesNumber = 1
    ds.AcqusitionNumber = 1
    
    ds.RescaleSlope = "1"
    ds.RescaleIntercept = "-1000"
    ds.RescaleType  = "HU"
    ds.WindowCenter = "-600"
    ds.WindowWidth = "1500"    
    ds.InstanceNumber = index


    spacings = meta_data['Spacing']
    spacings[2] /= opt['scale'] 
    # orientations = meta_data['Orientation']
    # origin = meta_data['Origin']  

    # ds.ImagePositionPatient = origin       
    # ds.SliceLocation = origin[2] - (index - 1) * spacings[2]   
    # ds.ImageOrientationPatient = orientations[:6]
    ds.SliceThickness = spacings[2] + 1e-10 
    ds.PixelSpacing = [spacings[0], spacings[1]]
    ds.PixelData = pixel_array.tobytes()
    ds.save_as(fullpath)
    

####################
# metric
####################
def calculate_psnr(img1, img2):
    # img1 and img2 have range [0, 1]
    img1 = img1.astype(np.float64)
    img2 = img2.astype(np.float64)
    mse = np.mean((img1 - img2)**2)
    if mse == 0:
        return float('inf')
    return 20 * math.log10(255.0 / math.sqrt(mse))


def ssim(img1, img2):
    C1 = (0.01 * 255)**2
    C2 = (0.03 * 255)**2
    # C1 = (0.01)**2
    # C2 = (0.03)**2
    img1 = img1.astype(np.float64)
    img2 = img2.astype(np.float64)
    kernel = cv2.getGaussianKernel(11, 1.5)
    window = np.outer(kernel, kernel.transpose())

    mu1 = cv2.filter2D(img1, -1, window)[5:-5, 5:-5]  # valid
    mu2 = cv2.filter2D(img2, -1, window)[5:-5, 5:-5]
    mu1_sq = mu1**2
    mu2_sq = mu2**2
    mu1_mu2 = mu1 * mu2
    sigma1_sq = cv2.filter2D(img1**2, -1, window)[5:-5, 5:-5] - mu1_sq
    sigma2_sq = cv2.filter2D(img2**2, -1, window)[5:-5, 5:-5] - mu2_sq
    sigma12 = cv2.filter2D(img1 * img2, -1, window)[5:-5, 5:-5] - mu1_mu2

    ssim_map = ((2 * mu1_mu2 + C1) * (2 * sigma12 + C2)) / ((mu1_sq + mu2_sq + C1) *
                                                            (sigma1_sq + sigma2_sq + C2))
    return ssim_map.mean()


def calculate_ssim(img1, img2):
    '''calculate SSIM
    the same outputs as MATLAB's
    img1, img2: [0, 255]
    '''
    if not img1.shape == img2.shape:
        raise ValueError('Input images must have the same dimensions.')
    if img1.ndim == 2:
        return ssim(img1, img2)
    elif img1.ndim == 3:
        if img1.shape[2] == 3:
            ssims = []
            for i in range(3):
                ssims.append(ssim(img1, img2))
            return np.array(ssims).mean()
        elif img1.shape[2] == 1:
            return ssim(np.squeeze(img1), np.squeeze(img2))
    else:
        raise ValueError('Wrong input image dimensions.')


def getIoU(y_true, y_pred):
    y_true_f = y_true.flatten()
    y_pred_f = y_pred.flatten()
    intersection = np.logical_and(y_true_f, y_pred_f).sum()
    union = np.logical_or(y_true_f, y_pred_f).sum()
    return (intersection + 1) * 1. / (union + 1)

"""
Creates 'pdist' model, which we need to calcuate 'pdist' metric during validation
"""
def create_pdist_model(network='vgg', use_gpu=True):
    '''
    percetual similarity metric
    https://github.com/richzhang/PerceptualSimilarity
    :param network can be vgg or alex
    '''
    print("Initializing pdist model")
    print("-"*40)
    # Initializing the model
    model = dm.DistModel()
    model.initialize(model='net-lin', net=network, use_gpu=use_gpu)
    return model


def calculate_pdist(model, img1, img2):
    # expand channel to 3 dim for vgg
    img1 = np.expand_dims(img1, axis=2)
    img1 = np.repeat(img1, 3, axis=2)
    img2 = np.expand_dims(img2, axis=2)
    img2 = np.repeat(img2, 3, axis=2)
    #  image from [-1,1]
    img1 = img2tensor(img1)
    img2 = img2tensor(img2)

    # Compute distance
    with torch.no_grad():
        dist = model.forward(img1, img2)
    return dist[0]

def cvt_int(self, inputString):
    find_digit = re.search(r'\d', inputString)
    reformat_num = int(float(inputString[find_digit.start():]))
    reformat_str = inputString[:find_digit.start()] + str(reformat_num)
    return reformat_str

def load_dicom(folder_location):
    '''
    :function 
    Load dicom images in one folder, give them correct ordering. 
    :returns
    A 3D volume and pixel spacings 
    '''
    # retrieve names of all dcm files
    dicom_names = glob.glob(os.path.join(folder_location,'*.dcm'))
    spacings = [] 
    # load dcm files 
    files = []
    for fname in dicom_names:
        files.append(pydicom.dcmread(fname))

    # skip files with no SliceLocation 
    slices = []
    skipcount = 0
    for f in files:
        if hasattr(f, 'SliceLocation'):
            slices.append(f)
        else:
            skipcount = skipcount + 1

    # ensure slices are in the correct order (sorted)
    slices = sorted(slices, key=lambda s: s.SliceLocation)

    # pixel spacings, assuming all slices are the same
    ps = [0, 0, 0]
    ps[0], ps[1] = float(slices[0].PixelSpacing[0]), float(slices[0].PixelSpacing[1])
    cosines0 = slices[0].ImageOrientationPatient
    iip0 = slices[0].ImagePositionPatient
    normal0 = [cosines0[1]*cosines0[5] - cosines0[2]*cosines0[4], 
                cosines0[2]*cosines0[3] - cosines0[0]*cosines0[5],
                cosines0[0]*cosines0[4] - cosines0[1]*cosines0[3]]
    dist0 = 0 
    for i in range(3): 
        dist0 += normal0[i]*iip0[i]
    
    iip1 = slices[1].ImagePositionPatient
    dist1 = 0 
    for i in range(3): 
        dist1 += normal0[i]*iip1[i]

    # subtract and take abs to obtain pixel spacing in z direction 
    ps[2] = abs(dist1 - dist0)

    # create 3D volume in [D,H,W]
    # img_shape = list(slices[0].pixel_array.shape)
    # img_shape.append(len(slices))
    img_shape = [len(slices), slices[0].pixel_array.shape[0], slices[0].pixel_array.shape[1]]
    vol = np.zeros(img_shape)

    # fill 3D array with the images from the files
    for i, s in enumerate(slices):
        img2d = s.pixel_array
        vol[i,:,:] = img2d
    
    # generate id 
    patient_id = slices[0].PatientID.replace("_", "-")
    timepoint = slices[0].StudyDate
    series_instance_uid = slices[0].SeriesInstanceUID
    modality = slices[0].Modality
    volume_id = f"{patient_id}_{timepoint}_{series_instance_uid}_{modality}"
    # print(volume_id)
    return vol, ps, volume_id


### options 
"""
Converts the 'OrderedDict()'to NoneDict, which return None for missing key
"""
def dict_to_nonedict(opt):
    if isinstance(opt, dict):
        new_opt = dict()
        for key, sub_opt in opt.items():
            new_opt[key] = dict_to_nonedict(sub_opt)
        return NoneDict(**new_opt)
    elif isinstance(opt, list):
        return [dict_to_nonedict(sub_opt) for sub_opt in opt]
    else:
        return opt

"""
Converts dictionary into readable string
"""
def dict2str(opt, indent_l=1):
    msg = ''
    for k, v in opt.items():
        if isinstance(v, dict):
            msg += ' ' * (indent_l * 2) + k + ':[\n'
            msg += dict2str(v, indent_l + 1)
            msg += ' ' * (indent_l * 2) + ']\n'
        else:
            msg += ' ' * (indent_l * 2) + k + ': ' + str(v) + '\n'
    return msg


class NoneDict(dict):
    def __missing__(self, key):
        return None

"""
Returns an object with a run_test function (for non_DL methods such as bm3d). 
"""
class PythonFunctionWrapper:
    def __init__(self, python_function):
        self.python_function = python_function

    def run_test(self, input_data):
        output = self.python_function(input_data)
        return output

def rescale_image(image, min_value, max_value):
    # Find the minimum and maximum values of the image
    current_min = np.min(image)
    current_max = np.max(image)

    # Rescale the image to the desired range
    rescaled_image = (image - current_min) * (max_value - min_value) / (current_max - current_min) + min_value

    return rescaled_image

def resize_image(data, size=(52,52,52)):
    """
    Resize the data to the input size
    """ 
    # [height, width, depth] = data.shape
    [depth, height, width] = data.shape 
    # scale = [self.input_D*1.0/depth, self.input_H*1.0/height, self.input_W*1.0/width] 
    target_depth, target_height, target_width = size 
    scale = [target_depth*1.0/depth, target_height*1.0/height, target_width*1.0/width] 
    data = ndimage.interpolation.zoom(data, scale, order=0)
    return data

def extract_roi(image, centroid): 
    """
    image: 3D numpy array in zyx 
    centorid: tuple (z, y, x) 
    Return a ROI of the input image centered at (z,y,x) and size (52,52,52)
    """
    # if isinstance(centroid, torch.Tensor):
    #     z, y, x = centroid.int()
    # else: 
    #     z, y, x = centroid

    z, y, x = centroid

    # Extract the bounding box of shape [52,52,52] around the centroid
    z_min, z_max = max(0, z - 26), min(image.shape[0], z + 26)
    y_min, y_max = max(0, y - 26), min(image.shape[1], y + 26)
    x_min, x_max = max(0, x - 26), min(image.shape[2], x + 26)
    # add zero padding if expected ROI lies partially out of the image                 
    roi = image[z_min:z_max, y_min:y_max, x_min:x_max]
    # pad_dims = [0, 52-(x_max-x_min), 0, 52-(y_max-y_min), 0, 52-(z_max-z_min)]
    # roi = F.pad(roi, pad_dims, mode='constant', value=0)
    pad_dims = ((0, 52-(z_max-z_min)), (0, 52-(y_max-y_min)), (0, 52-(x_max-x_min)))
    roi = np.pad(roi, pad_width=pad_dims, mode='constant', constant_values=0)
    
    return roi 

def extract_roi_with_diameter(image, centroid, diameter):
    z, y, x = centroid

    # Calculate starting and ending indices for each dimension
    start_z = max(0, z - diameter // 2)
    end_z = min(image.shape[0], z + diameter // 2 + 1)
    start_y = max(0, y - diameter // 2)
    end_y = min(image.shape[1], y + diameter // 2 + 1)
    start_x = max(0, x - diameter // 2)
    end_x = min(image.shape[2], x + diameter // 2 + 1)

    # Slice the image and extract the ROI
    roi = image[start_z:end_z, start_y:end_y, start_x:end_x]

    return roi

def convert_numpy_to_list(data):
    if isinstance(data, np.ndarray):
        return data.tolist()
    elif isinstance(data, dict):
        return {key: convert_numpy_to_list(value) for key, value in data.items()}
    elif isinstance(data, list):
        return [convert_numpy_to_list(item) for item in data]
    else:
        return data



import os
import math
import sys
import matplotlib
matplotlib.use('agg')
import matplotlib.pyplot as plt
from matplotlib.ticker import ScalarFormatter,LogFormatter,StrMethodFormatter,FixedFormatter
import sklearn.metrics as skl_metrics
import numpy as np

from evaluationScript.NoduleFinding import NoduleFinding

from evaluationScript.tools import csvTools
# matplotlib.rc('xtick', labelsize=18)
# matplotlib.rc('ytick', labelsize=18)
font = {'family' : 'normal',
        'size'   : 17}

matplotlib.rc('font', **font)
# Evaluation settings
bPerformBootstrapping = False
bNumberOfBootstrapSamples = 1000
bOtherNodulesAsIrrelevant = True
bConfidence = 0.95

seriesuid_label = 'pid'
coordX_label = 'center_x'
coordY_label = 'center_y'
coordZ_label = 'center_z'
diameter_mm_label = 'diameter'
CADProbability_label = 'probability'
xmin_label = 'xmin'
xmax_label = 'xmax'
ymin_label = 'ymin'
ymax_label = 'ymax'
zmin_label = 'zmin'
zmax_label = 'zmax'

# plot settings
FROC_minX = 0.125 # Mininum value of x-axis of FROC curve
FROC_maxX = 8 # Maximum value of x-axis of FROC curve
bLogPlot = True

def generateBootstrapSet(scanToCandidatesDict, FROCImList):
    '''
    Generates bootstrapped version of set
    '''
    imageLen = FROCImList.shape[0]
    
    # get a random list of images using sampling with replacement
    rand_index_im   = np.random.randint(imageLen, size=imageLen)
    FROCImList_rand = FROCImList[rand_index_im]
    
    # get a new list of candidates
    candidatesExists = False
    for im in FROCImList_rand:
        if im not in scanToCandidatesDict:
            continue
        
        if not candidatesExists:
            candidates = np.copy(scanToCandidatesDict[im])
            candidatesExists = True
        else:
            candidates = np.concatenate((candidates,scanToCandidatesDict[im]),axis = 1)

    return candidates

def compute_mean_ci(interp_sens, confidence = 0.95):
    sens_mean = np.zeros((interp_sens.shape[1]),dtype = 'float32')
    sens_lb   = np.zeros((interp_sens.shape[1]),dtype = 'float32')
    sens_up   = np.zeros((interp_sens.shape[1]),dtype = 'float32')
    
    Pz = (1.0-confidence)/2.0
    print(interp_sens.shape)
    for i in range(interp_sens.shape[1]):
        # get sorted vector
        vec = interp_sens[:,i]
        vec.sort()

        sens_mean[i] = np.average(vec)
        sens_lb[i] = vec[int(math.floor(Pz*len(vec)))]
        sens_up[i] = vec[int(math.floor((1.0-Pz)*len(vec)))]

    return sens_mean,sens_lb,sens_up

def computeFROC_bootstrap(FROCGTList,FROCProbList,FPDivisorList,FROCImList,excludeList,numberOfBootstrapSamples=1000, confidence = 0.95):

    set1 = np.concatenate(([FROCGTList], [FROCProbList], [excludeList]), axis=0)
    
    fps_lists = []
    sens_lists = []
    thresholds_lists = []
    
    FPDivisorList_np = np.asarray(FPDivisorList)
    FROCImList_np = np.asarray(FROCImList)
    
    # Make a dict with all candidates of all scans
    scanToCandidatesDict = {}
    for i in range(len(FPDivisorList_np)):
        seriesuid = FPDivisorList_np[i]
        candidate = set1[:,i:i+1]

        if seriesuid not in scanToCandidatesDict:
            scanToCandidatesDict[seriesuid] = np.copy(candidate)
        else:
            scanToCandidatesDict[seriesuid] = np.concatenate((scanToCandidatesDict[seriesuid],candidate),axis = 1)

    for i in range(numberOfBootstrapSamples):
        print('computing FROC: bootstrap %d/%d' % (i,numberOfBootstrapSamples))
        # Generate a bootstrapped set
        btpsamp = generateBootstrapSet(scanToCandidatesDict,FROCImList_np)
        fps, sens, thresholds = computeFROC(btpsamp[0,:],btpsamp[1,:],len(FROCImList_np),btpsamp[2,:])
    
        fps_lists.append(fps)
        sens_lists.append(sens)
        thresholds_lists.append(thresholds)

    # compute statistic
    all_fps = np.linspace(FROC_minX, FROC_maxX, num=10000)
    
    # Then interpolate all FROC curves at this points
    interp_sens = np.zeros((numberOfBootstrapSamples,len(all_fps)), dtype = 'float32')
    for i in range(numberOfBootstrapSamples):
        interp_sens[i,:] = np.interp(all_fps, fps_lists[i], sens_lists[i])
    
    # compute mean and CI
    sens_mean,sens_lb,sens_up = compute_mean_ci(interp_sens, confidence = confidence)

    return all_fps, sens_mean, sens_lb, sens_up

def computeFROC(FROCGTList, FROCProbList, totalNumberOfImages, excludeList):
    # Remove excluded candidates
    FROCGTList_local = []
    FROCProbList_local = []
    for i in range(len(excludeList)):
        if excludeList[i] == False:
            FROCGTList_local.append(FROCGTList[i])
            FROCProbList_local.append(FROCProbList[i])
    
    numberOfDetectedLesions = sum(FROCGTList_local)
    totalNumberOfLesions = sum(FROCGTList)
    totalNumberOfCandidates = len(FROCProbList_local)
    fpr, tpr, thresholds = skl_metrics.roc_curve(FROCGTList_local, FROCProbList_local)
    if sum(FROCGTList) == len(FROCGTList): # Handle border case when there are no false positives and ROC analysis give nan values.
      print("WARNING, this system has no false positives..")
      fps = np.zeros(len(fpr))
    else:
      fps = fpr * (totalNumberOfCandidates - numberOfDetectedLesions) / totalNumberOfImages
    sens = (tpr * numberOfDetectedLesions) / totalNumberOfLesions
    return fps, sens, thresholds

def evaluateCAD(seriesUIDs, results_filename, outputDir, allNodules, CADSystemName, maxNumberOfCADMarks=-1,
                performBootstrapping=False,numberOfBootstrapSamples=1000,confidence = 0.95):
    '''
    function to evaluate a CAD algorithm
    @param seriesUIDs: list of the seriesUIDs of the cases to be processed
    @param results_filename: file with results
    @param outputDir: output directory
    @param allNodules: dictionary with all nodule annotations of all cases, keys of the dictionary are the seriesuids
    @param CADSystemName: name of the CAD system, to be used in filenames and on FROC curve
    '''

    nodOutputfile = open(os.path.join(outputDir,'CADAnalysis.txt'),'w')
    nodOutputfile.write("\n")
    nodOutputfile.write((60 * "*") + "\n")
    nodOutputfile.write("CAD Analysis: %s\n" % CADSystemName)
    nodOutputfile.write((60 * "*") + "\n")
    nodOutputfile.write("\n")

    results = csvTools.readCSV(results_filename)
    allCandsCAD = {}
    
    # collect candidates from result file
    for seriesuid in seriesUIDs:
        nodules = {}
        header = results[0]
        i = 0
        for result in results[1:]:
            nodule_seriesuid = result[header.index(seriesuid_label)]
            if seriesuid == nodule_seriesuid:
                nodule = getNodule(result, header)
                nodule.candidateID = i
                nodules[nodule.candidateID] = nodule
                i += 1
        if (maxNumberOfCADMarks > 0):
            # number of CAD marks, only keep must suspicous marks
            if len(nodules.keys()) > maxNumberOfCADMarks:
                # make a list of all probabilities
                probs = []
                for keytemp, noduletemp in nodules.items():
                    probs.append(float(noduletemp.CADprobability))
                probs.sort(reverse=True) # sort from large to small
                probThreshold = probs[maxNumberOfCADMarks] - 1e-8
                nodules2 = {}
                nrNodules2 = 0
                for keytemp, noduletemp in nodules.items():
                    if nrNodules2 >= maxNumberOfCADMarks:
                        break
                    if float(noduletemp.CADprobability) > probThreshold:
                        nodules2[keytemp] = noduletemp
                        nrNodules2 += 1

                nodules = nodules2
        
        # print('adding candidates: ' + seriesuid)
        allCandsCAD[seriesuid] = nodules
    
    # open output files
    nodNoCandFile = open(os.path.join(outputDir, "nodulesWithoutCandidate_%s.txt" % CADSystemName), 'w')
    
    # --- iterate over all cases (seriesUIDs) and determine how
    # often a nodule annotation is not covered by a candidate

    # initialize some variables to be used in the loop
    candTPs = 0
    candFPs = 0
    candFNs = 0
    candTNs = 0
    totalNumberOfCands = 0
    totalNumberOfNodules = 0
    doubleCandidatesIgnored = 0
    irrelevantCandidates = 0
    minProbValue = -1000000000.0 # minimum value of a float
    FROCGTList = []
    FROCProbList = []
    FPDivisorList = []
    excludeList = []
    FROCtoNoduleMap = []
    ignoredCADMarksList = []

    # -- loop over the cases
    print('======= evaluation CAD ========')
    for seriesuid in seriesUIDs:
        # get the candidates for this case
        try:
            candidates = allCandsCAD[seriesuid]
        except KeyError:
            candidates = {}
        # print('UID:', seriesuid)
        # print('num candidates:', len(candidates))
        # add to the total number of candidates
        totalNumberOfCands += len(candidates.keys())
        # make a copy in which items will be deleted
        candidates2 = candidates.copy()
        # print('second candidate keys:', candidates2.keys())
        
        # get the nodule annotations on this case
        try:
            noduleAnnots = allNodules[seriesuid]
        except KeyError:
            noduleAnnots = []

        # print('GT nodule exists:', len(noduleAnnots))

        # - loop over the nodule annotations
        # print('Looping over GT nodule')
        for nod_idx, noduleAnnot in enumerate(noduleAnnots):
            # increment the number of nodules
            if noduleAnnot.state == "Included":
                totalNumberOfNodules += 1

            # print('nodule :', nod_idx+1)
            x = float(noduleAnnot.coordX)
            y = float(noduleAnnot.coordY)
            z = float(noduleAnnot.coordZ)
            diameter_mm = float(noduleAnnot.diameter_mm)
            
            nod_xmin = float(noduleAnnot.xmin)
            nod_xmax = float(noduleAnnot.xmax)
            nod_ymin = float(noduleAnnot.ymin)
            nod_ymax = float(noduleAnnot.ymax)
            nod_zmin = float(noduleAnnot.zmin)
            nod_zmax = float(noduleAnnot.zmax)

            # 2. Check if the nodule annotation is covered by a candidate
            # A nodule is marked as detected when the center of mass of the candidate is within a distance R of
            # the center of the nodule. In order to ensure that the CAD mark is displayed within the nodule on the
            # CT scan, we set R to be the radius of the nodule size.
            """
            diameter = float(noduleAnnot.diameter_mm)
            if diameter < 0.0:
              diameter = 10.0
            radiusSquared = pow((diameter / 2.0), 2.0)
            """

            found = False
            noduleMatches = []
            # print('for nod {}, looping over candidates:'.format(nod_idx+1))
            for key, candidate in candidates.items():
                """
                x2 = float(candidate.coordX)
                y2 = float(candidate.coordY)
                z2 = float(candidate.coordZ)
                dist = math.pow(x - x2, 2.) + math.pow(y - y2, 2.) + math.pow(z - z2, 2.)
                """
                cand_x = float(candidate.coordX)
                cand_y = float(candidate.coordY)
                cand_z = float(candidate.coordZ)
                
                # create a box equal to size of diam in x,y,z and check if predicted centroid lies within that box
                if (x+diameter_mm >= cand_x >= x-diameter_mm) and (y+diameter_mm >= cand_y >= y-diameter_mm) and (z+diameter_mm >= cand_z >= z-diameter_mm):
                # if (nod_xmax >= cand_x >= nod_xmin) and (nod_ymax >= cand_y >= nod_ymin) and (nod_zmax >= cand_z >= nod_zmin):
                    # print('candidate match GT for candidate key: {} | XYZ:[{},{},{}]'.format(key, candidate.coordX, candidate.coordY, candidate.coordZ))
                    if (noduleAnnot.state == "Included"):
                        found = True
                        noduleMatches.append(candidate)
                        if key not in candidates2.keys():
                            # note: if we go by detecting whether cand (x,y,z) lies within GT bbox, if two GT nodules are close to each other,
                            # same candidate can cover both GT nodule (meaning same x,y,z can be within range of two GT bbox) and in that case,
                            # it will print warning
                            print("This is strange: CAD mark %s detected two nodules! Check for overlapping nodule annotations, SeriesUID: %s, nodule Annot ID: %s" % (str(candidate.id), seriesuid, str(noduleAnnot.id)))
                        else:
                            # print('deleting key {} for candidates2'.format(key))
                            del candidates2[key]
                    elif (noduleAnnot.state == "Excluded"): # an excluded nodule
                        if bOtherNodulesAsIrrelevant: #    delete marks on excluded nodules so they don't count as false positives
                            if key in candidates2.keys():
                                irrelevantCandidates += 1
                                ignoredCADMarksList.append("%s,%s,%s,%s,%s,%s,%.9f" % (seriesuid, -1, candidate.coordX, candidate.coordY, candidate.coordZ, str(candidate.id), float(candidate.CADprobability)))
                                del candidates2[key]
                
                """
                if dist < radiusSquared:
                    if (noduleAnnot.state == "Included"):
                        found = True
                        noduleMatches.append(candidate)
                        if key not in candidates2.keys():
                            print("This is strange: CAD mark %s detected two nodules! Check for overlapping nodule annotations, SeriesUID: %s, nodule Annot ID: %s" % (str(candidate.id), seriesuid, str(noduleAnnot.id)))
                        else:
                            del candidates2[key]
                    elif (noduleAnnot.state == "Excluded"): # an excluded nodule
                        if bOtherNodulesAsIrrelevant: #    delete marks on excluded nodules so they don't count as false positives
                            if key in candidates2.keys():
                                irrelevantCandidates += 1
                                ignoredCADMarksList.append("%s,%s,%s,%s,%s,%s,%.9f" % (seriesuid, -1, candidate.coordX, candidate.coordY, candidate.coordZ, str(candidate.id), float(candidate.CADprobability)))
                                del candidates2[key]
                """


            if len(noduleMatches) > 1: # double detection
                doubleCandidatesIgnored += (len(noduleMatches) - 1)
            if noduleAnnot.state == "Included":
                # only include it for FROC analysis if it is included
                # otherwise, the candidate will not be counted as FP, but ignored in the
                # analysis since it has been deleted from the nodules2 vector of candidates
                if found == True:
                    # append the sample with the highest probability for the FROC analysis
                    maxProb = None
                    for idx in range(len(noduleMatches)):
                        candidate = noduleMatches[idx]
                        if (maxProb is None) or (float(candidate.CADprobability) > maxProb):
                            maxProb = float(candidate.CADprobability)

                    FROCGTList.append(1.0)
                    FROCProbList.append(float(maxProb))
                    FPDivisorList.append(seriesuid)
                    excludeList.append(False)
                    FROCtoNoduleMap.append("%s,%s,%s,%s,%s,%.9f,%s,%.9f" % (seriesuid, noduleAnnot.id, noduleAnnot.coordX, noduleAnnot.coordY, noduleAnnot.coordZ, float(noduleAnnot.diameter_mm), str(candidate.id), float(candidate.CADprobability)))
                    candTPs += 1
                else:
                    candFNs += 1
                    # append a positive sample with the lowest probability, such that this is added in the FROC analysis
                    FROCGTList.append(1.0)
                    FROCProbList.append(minProbValue)
                    FPDivisorList.append(seriesuid)
                    excludeList.append(True)
                    FROCtoNoduleMap.append("%s,%s,%s,%s,%s,%.9f,%s,%s" % (seriesuid, noduleAnnot.id, noduleAnnot.coordX, noduleAnnot.coordY, noduleAnnot.coordZ, float(noduleAnnot.diameter_mm), int(-1), "NA"))
                    nodNoCandFile.write("%s,%s,%s,%s,%s,%.9f,%s\n" % (seriesuid, noduleAnnot.id, noduleAnnot.coordX, noduleAnnot.coordY, noduleAnnot.coordZ, float(noduleAnnot.diameter_mm), str(-1)))

        # add all false positives to the vectors
        for key, candidate3 in candidates2.items():
            candFPs += 1
            FROCGTList.append(0.0)
            FROCProbList.append(float(candidate3.CADprobability))
            FPDivisorList.append(seriesuid)
            excludeList.append(False)
            FROCtoNoduleMap.append("%s,%s,%s,%s,%s,%s,%.9f" % (seriesuid, -1, candidate3.coordX, candidate3.coordY, candidate3.coordZ, str(candidate3.id), float(candidate3.CADprobability)))

    if not (len(FROCGTList) == len(FROCProbList) and len(FROCGTList) == len(FPDivisorList) and len(FROCGTList) == len(FROCtoNoduleMap) and len(FROCGTList) == len(excludeList)):
        nodOutputfile.write("Length of FROC vectors not the same, this should never happen! Aborting..\n")

    nodOutputfile.write("Candidate detection results:\n")
    nodOutputfile.write("    True positives: %d\n" % candTPs)
    nodOutputfile.write("    False positives: %d\n" % candFPs)
    nodOutputfile.write("    False negatives: %d\n" % candFNs)
    nodOutputfile.write("    True negatives: %d\n" % candTNs)
    nodOutputfile.write("    Total number of candidates: %d\n" % totalNumberOfCands)
    nodOutputfile.write("    Total number of nodules: %d\n" % totalNumberOfNodules)

    nodOutputfile.write("    Ignored candidates on excluded nodules: %d\n" % irrelevantCandidates)
    nodOutputfile.write("    Ignored candidates which were double detections on a nodule: %d\n" % doubleCandidatesIgnored)
    if int(totalNumberOfNodules) == 0:
        nodOutputfile.write("    Sensitivity: 0.0\n")
    else:
        nodOutputfile.write("    Sensitivity: %.9f\n" % (float(candTPs) / float(totalNumberOfNodules)))
    nodOutputfile.write("    Average number of candidates per scan: %.9f\n" % (float(totalNumberOfCands) / float(len(seriesUIDs))))

    # compute FROC
    fps, sens, thresholds = computeFROC(FROCGTList,FROCProbList,len(seriesUIDs),excludeList)
    if performBootstrapping:
        fps_bs_itp,sens_bs_mean,sens_bs_lb,sens_bs_up = computeFROC_bootstrap(FROCGTList,FROCProbList,FPDivisorList,seriesUIDs,excludeList,
                                                                  numberOfBootstrapSamples=numberOfBootstrapSamples, confidence = confidence)
        
    # Write FROC curve
    with open(os.path.join(outputDir, "froc_%s.txt" % CADSystemName), 'w') as f:
        for i in range(len(sens)):
            f.write("%.9f,%.9f,%.9f\n" % (fps[i], sens[i], thresholds[i]))
    
    # Write FROC vectors to disk as well
    with open(os.path.join(outputDir, "froc_gt_prob_vectors_%s.csv" % CADSystemName), 'w') as f:
        for i in range(len(FROCGTList)):
            f.write("%d,%.9f\n" % (FROCGTList[i], FROCProbList[i]))

    fps_itp = np.linspace(FROC_minX, FROC_maxX, num=10001)
    sens_itp = np.interp(fps_itp, fps, sens)
    frvvlu = 0
    nxth = 0.125
    for fp, ss in zip(fps_itp, sens_itp):
        if abs(fp - nxth) < 3e-4:
            frvvlu += ss
            nxth *= 2
        if abs(nxth - 16) < 1e-5: break
    print(frvvlu/7, nxth)
    print(sens_itp[fps_itp==0.125]+sens_itp[fps_itp==0.25]+sens_itp[fps_itp==0.5]+sens_itp[fps_itp==1]+sens_itp[fps_itp==2]\
        +sens_itp[fps_itp==4]+sens_itp[fps_itp==8])
    if performBootstrapping:
        # Write mean, lower, and upper bound curves to disk
        with open(os.path.join(outputDir, "froc_%s_bootstrapping.csv" % CADSystemName), 'w') as f:
            f.write("FPrate,Sensivity[Mean],Sensivity[Lower bound],Sensivity[Upper bound]\n")
            for i in range(len(fps_bs_itp)):
                f.write("%.9f,%.9f,%.9f,%.9f\n" % (fps_bs_itp[i], sens_bs_mean[i], sens_bs_lb[i], sens_bs_up[i]))
    else:
        fps_bs_itp = None
        sens_bs_mean = None
        sens_bs_lb = None
        sens_bs_up = None

    # create FROC graphs
    if int(totalNumberOfNodules) > 0:
        graphTitle = str("")
        fig1 = plt.figure()
        ax = plt.gca()
        clr = 'b'
        plt.plot(fps_itp, sens_itp, color=clr, label="%s" % CADSystemName, lw=2)
        if performBootstrapping:
            plt.plot(fps_bs_itp, sens_bs_mean, color=clr, ls='--')
            plt.plot(fps_bs_itp, sens_bs_lb, color=clr, ls=':') # , label = "lb")
            plt.plot(fps_bs_itp, sens_bs_up, color=clr, ls=':') # , label = "ub")
            ax.fill_between(fps_bs_itp, sens_bs_lb, sens_bs_up, facecolor=clr, alpha=0.05)
        xmin = FROC_minX
        xmax = FROC_maxX
        plt.xlim(xmin, xmax)
        plt.ylim(0, 1)
        plt.xlabel('Average number of false positives per scan')
        plt.ylabel('Sensitivity')
        plt.legend(loc='lower right')
        plt.title('FROC performance - %s' % (CADSystemName))
        
        if bLogPlot:
            plt.xscale('log', basex=2)
            ax.xaxis.set_major_formatter(FixedFormatter([0.125,0.25,0.5,1,2,4,8]))
        
        # set your ticks manually
        ax.xaxis.set_ticks([0.125,0.25,0.5,1,2,4,8])
        # ax.yaxis.set_ticks(np.arange(0.5, 1, 0.1))
        ax.yaxis.set_ticks(np.arange(0, 1.1, 0.1))
        plt.grid(b=True, which='both')
        plt.tight_layout()
        plt.savefig(os.path.join(outputDir, "froc_%s.png" % CADSystemName), bbox_inches=0, dpi=300)

    def find_nearest(array,value):
        idx = np.where((fps - value) == (fps - value)[(fps - value) >= 0].min())[0][-1]
        return idx
    thres = [0.125,0.25,0.5,1,2,4,8]
    sen = []
    print('FROC at points: ', thres)
    for th in thres:
        print('fps: ', th, ', sensitivity: ', sens[find_nearest(fps, th)])
        sen.append(sens[find_nearest(fps, th)])
    print('=============================================\naverage FROC: ', np.mean(sen))
    out_dict = {'CM':{'TP':candTPs, 'FP':candFPs, 'TN':candTNs, 'FN':candFNs}, 'FROC_FPs':thres, 'FROC_sens':sen,
                    'FROC_gph':fps, 'Sens_gph': sens, 'FROC_gph_itp':fps_itp, 'Sens_gph_itp':sens_itp,
                    'overall_sens': (float(candTPs) / float(totalNumberOfNodules))
               }
    # return (fps, sens, thresholds, fps_bs_itp, sens_bs_mean, sens_bs_lb, sens_bs_up)
    return out_dict


def getNodule(annotation, header, state = ""):
    nodule = NoduleFinding()
    nodule.coordX = annotation[header.index(coordX_label)]
    nodule.coordY = annotation[header.index(coordY_label)]
    nodule.coordZ = annotation[header.index(coordZ_label)]
    if xmin_label in header:
        nodule.xmin = annotation[header.index(xmin_label)]
        nodule.xmax = annotation[header.index(xmax_label)]
        nodule.ymin = annotation[header.index(ymin_label)]
        nodule.ymax = annotation[header.index(ymax_label)]
        nodule.zmin = annotation[header.index(zmin_label)]
        nodule.zmax = annotation[header.index(zmax_label)]
    
    if diameter_mm_label in header:
        nodule.diameter_mm = annotation[header.index(diameter_mm_label)]
    if CADProbability_label in header:
        nodule.CADprobability = annotation[header.index(CADProbability_label)]
    if not state == "":
        nodule.state = state

    return nodule
    
def collectNoduleAnnotations(annotations, seriesUIDs):
    allNodules = {}
    noduleCount = 0
    noduleCountTotal = 0
    for seriesuid in seriesUIDs:
        nodules = []
        numberOfIncludedNodules = 0
        
        # add included findings
        header = annotations[0]
        for annotation in annotations[1:]:
            nodule_seriesuid = annotation[header.index(seriesuid_label)]
            
            if seriesuid == nodule_seriesuid:
                nodule = getNodule(annotation, header, state = "Included")
                nodules.append(nodule)
                numberOfIncludedNodules += 1
            """
            if int(seriesuid) == int(nodule_seriesuid):
                nodule = getNodule(annotation, header, state = "Included")
                nodules.append(nodule)
                numberOfIncludedNodules += 1
            """

        allNodules[seriesuid] = nodules
        noduleCount      += numberOfIncludedNodules
        noduleCountTotal += len(nodules)
    
    print('Total number of included nodule annotations: ' + str(noduleCount))
    print('Total number of nodule annotations: ' + str(noduleCountTotal))
    return allNodules
    
    
def collect(annotations_filename, valpath):
    annotations = csvTools.readCSV(annotations_filename)
    with open(valpath, "r") as f:
        testfilelist = f.read().splitlines()
    seriesUIDs = []
    for seriesUID in testfilelist:
        seriesUIDs.append(seriesUID)
    print('Num of seriesUIDs collected:', len(seriesUIDs))
    print(seriesUIDs)
    print('-'*40)
    allNodules = collectNoduleAnnotations(annotations, seriesUIDs)
    return (allNodules, seriesUIDs)

    
def noduleCADEvaluation(annotations_filename,results_filename,valpath,outputDir, return_dict=False):
    '''
    function to load annotations and evaluate a CAD algorithm
    @param annotations_filename: list of annotations
    @param annotations_excluded_filename: list of annotations that are excluded from analysis
    @param seriesuids_filename: list of CT images in seriesuids
    @param results_filename: list of CAD marks with probabilities
    @param outputDir: output directory
    '''
    print('GT annot:', annotations_filename)
    print('Prediction annot:', results_filename)
    print('UIDS:', valpath)
    print('-'*40)
    print('collecting nodules....')
    (allNodules, seriesUIDs) = collect(annotations_filename, valpath)
    output_dict = evaluateCAD(seriesUIDs, results_filename, outputDir, allNodules,
                os.path.splitext(os.path.basename(results_filename))[0],
                maxNumberOfCADMarks=300, performBootstrapping=bPerformBootstrapping,
                numberOfBootstrapSamples=bNumberOfBootstrapSamples, confidence=bConfidence)

    if return_dict:
        return output_dict


if __name__ == '__main__':
    annotations_filename = '/home/mj/dataset/nodule_data/data/nodule_process/process_zoom_validate/label_3d.csv'
    val_path = '/home/mj/dataset/nodule_data/data/nodule_process/process_zoom_validate/val.txt'

    results_filename = '../results/validate1/res/51/FROC/submission_rpn.csv'#3D Faster R-CNN - Res18.csv' #top5.csv'#
    noduleCADEvaluation(annotations_filename,
                        results_filename, val_path, os.path.join('../results/validate1/res/51/FROC', 'rpn'))
    print("Finished!")


import pandas as pd
import os
import numpy as np


def _populate_value(incoming_lines, file_substr, data_dict):
    for line in incoming_lines:
        metric = line.split(' {} '.format(file_substr))[-1].split(' for ')[0].upper()
        view = line.split(' in ')[-1].split(' view' )[0].upper()
        value = float(line.split(' view :')[-1])
        d_place = float("{0:.4f}".format(value))
        data_dict[metric][view].append(d_place)
    return data_dict

def _save_csv(data_dict, num_sub, merge_folds, path_to_save):
    for each_model in data_dict:
        model_data = data_dict[each_model]
        for each_metric in model_data:
            metric_data = model_data[each_metric]
            for each_view in metric_data:
                view_data = np.array(metric_data[each_view])
                if merge_folds:
                    csv_dict = {'Subject':np.arange(1, 5+1, 1)} # 5 folds
                    # csv_dict['{}_out'.format(each_model)] = np.mean(view_data, axis=0)
                    csv_dict['{}_out'.format(each_model)] = np.std(view_data, axis=1)
                else:
                    csv_dict = {'Subject':np.arange(1, num_sub+1, 1)}
                    csv_dict['{}_out'.format(each_model)] = view_data
                    print('csv ditc:', csv_dict)
                file_name = "SD_across_Sub_{}_{}_{}.csv".format(each_model, each_metric, each_view)
                file_path = os.path.join(path_to_save, file_name)
                df = pd.DataFrame.from_dict(csv_dict) 
                df.to_csv (file_path, index=False, header=True)
    print('ALL Files Created!')

def get_avg_values(log_files, file_substr, start_from, num_sub, num_lines_to_skip, merge_folds):
    # scan all files
    master_dict = {}
    for root, dirs, files in os.walk(log_files, topdown=True):
        if len(files) == 1 or len(files) == 2:
            for file in files:
                if file.endswith('.log') and file_substr in file:
                    file_path = os.path.join(root, file)
                    fold_name = root.split('/')[-1]
                    master_dict[fold_name] = {}
                    # read file for a specific fold/case
                    with open(file_path, 'r') as file:
                        lines = file.readlines()
                        master_dict[fold_name] = {  'PSNR':{'XY': [], 'XZ':[], 'YZ':[]},
                                                    'SSIM':{'XY': [], 'XZ':[], 'YZ':[]},
                                                    'PDIST':{'XY': [], 'XZ':[], 'YZ':[]}
                                                 }
                        start_lines_at, end_lines_at = 0, num_sub 
                        for index, counter in enumerate(range(0, len(lines), num_sub)):
                            if index == 0:
                                lines_of_interest = lines[start_lines_at:end_lines_at]
                            else:
                                lines_of_interest = lines[end_lines_at+num_lines_to_skip: end_lines_at+num_lines_to_skip+num_sub]
                                end_lines_at = end_lines_at + num_sub + num_lines_to_skip
                            # read values into a dictionary            
                            if len(lines_of_interest) != 0:
                                master_dict[fold_name] = _populate_value(lines_of_interest, file_substr, master_dict[fold_name])

    os.makedirs('../csv_files', exist_ok=True)
    # merge csv files before saving
    if merge_folds:
        per_fold_data = {}
        for each_fold in master_dict:
            model_name = each_fold.split('k2d100-')[-1].split('-L1')[0]
            if model_name not in per_fold_data:
                per_fold_data[model_name] = { 'PSNR':{'XY': [], 'XZ':[], 'YZ':[]},
                                              'SSIM':{'XY': [], 'XZ':[], 'YZ':[]},
                                              'PDIST':{'XY': [], 'XZ':[], 'YZ':[]}
                                            }
            # combine data for each fold
            per_fold_data[model_name]['PSNR']['XY'].append(master_dict[each_fold]['PSNR']['XY']) 
            per_fold_data[model_name]['PSNR']['XZ'].append(master_dict[each_fold]['PSNR']['XZ']) 
            per_fold_data[model_name]['PSNR']['YZ'].append(master_dict[each_fold]['PSNR']['YZ'])
            per_fold_data[model_name]['SSIM']['XY'].append(master_dict[each_fold]['SSIM']['XY']) 
            per_fold_data[model_name]['SSIM']['XZ'].append(master_dict[each_fold]['SSIM']['XZ']) 
            per_fold_data[model_name]['SSIM']['YZ'].append(master_dict[each_fold]['SSIM']['YZ'])
            per_fold_data[model_name]['PDIST']['XY'].append(master_dict[each_fold]['PDIST']['XY']) 
            per_fold_data[model_name]['PDIST']['XZ'].append(master_dict[each_fold]['PDIST']['XZ']) 
            per_fold_data[model_name]['PDIST']['YZ'].append(master_dict[each_fold]['PDIST']['YZ'])      
        _save_csv(per_fold_data, num_sub, merge_folds, '../csv_files')
    else:
        _save_csv(master_dict, num_sub, merge_folds, '../csv_files')


if __name__ == "__main__":
    # path_to_logs = '../../experiments/cv_logs'
    # get_avg_values(path_to_logs, file_substr='val', start_from=3611, num_sub=20, num_lines_to_skip=1, merge_folds=True)
    path_to_logs = '../../results_ucla'
    get_avg_values(path_to_logs, file_substr='test', start_from=0, num_sub=32, num_lines_to_skip=1, merge_folds=False)
    


import os
import numpy as np
import torch
import random

# Set seed
SEED = 35202
random.seed(SEED)
np.random.seed(SEED)
torch.manual_seed(SEED)
torch.cuda.manual_seed_all(SEED)


# Preprocessing using preserved HU in dilated part of mask
BASE = '/workspace/SANetwork/' # make sure you have the ending '/'  # /home/media/ssd/process_zoom/
data_config = {
    # directory for putting all preprocessed results for training to this path
    # 'preprocessed_data_dir': BASE + 'full',
    # 'preprocessed_data_dir': '/datasets/reference/k2_d100_st1',
    'preprocessed_data_dir': BASE + 'LUNA16_subset',
    'roi_names': ['nodule'],
    'crop_size': [128, 128, 128],
    'bbox_border': 8,
    'pad_value': 170,
    # 'jitter_range': [0, 0, 0],
}

def get_anchors(bases, aspect_ratios):
    anchors = []
    for b in bases:
        for asp in aspect_ratios:
            d, h, w = b * asp[0], b * asp[1], b * asp[2]
            anchors.append([d, h, w])

    return anchors

bases = [5, 10, 20, 30, 50]
aspect_ratios = [[1, 1, 1]]

net_config = {
    # Net configuration
    'anchors': get_anchors(bases, aspect_ratios),
    'chanel': 1,
    'crop_size': data_config['crop_size'],
    'stride': 4,
    'max_stride': 16,
    'num_neg': 800,
    'th_neg': 0.02,
    'th_pos_train': 0.5,
    'th_pos_val': 1,
    'num_hard': 3,
    'bound_size': 12,
    'blacklist': ['09184', '05181','07121','00805','08832'],

    'augtype': {'flip': False, 'rotate': False, 'scale': True, 'swap': False},
    'r_rand_crop': 0.,
    'pad_value': 170,

    # region proposal network configuration
    'rpn_train_bg_thresh_high': 0.02,
    'rpn_train_fg_thresh_low': 0.5,
    
    'rpn_train_nms_num': 300,
    'rpn_train_nms_pre_score_threshold': 0.5,
    'rpn_train_nms_overlap_threshold': 0.1,
    'rpn_test_nms_pre_score_threshold': 0.5,
    'rpn_test_nms_overlap_threshold': 0.1,

    # false positive reduction network configuration
    'num_class': len(data_config['roi_names']) + 1,
    'rcnn_crop_size': (7,7,7), # can be set smaller, should not affect much
    'rcnn_train_fg_thresh_low': 0.5,
    'rcnn_train_bg_thresh_high': 0.1,
    'rcnn_train_batch_size': 64,
    'rcnn_train_fg_fraction': 0.5,
    'rcnn_train_nms_pre_score_threshold': 0.5,
    'rcnn_train_nms_overlap_threshold': 0.1,
    'rcnn_test_nms_pre_score_threshold': 0.5, # 0.0
    'rcnn_test_nms_overlap_threshold': 0.8, # 0.1
    'box_reg_weight': [1., 1., 1., 1., 1., 1.]
}


def lr_shedule(epoch, init_lr=0.01, total=200):
    if epoch <= total * 0.5:
        lr = init_lr
    elif epoch <= total * 0.8:
        lr = 0.1 * init_lr
    else:
        lr = 0.01 * init_lr
    return lr

train_config = {
    'net': 'SANet',
    'batch_size': 16,
    'lr_schedule': lr_shedule,
    'optimizer': 'SGD',
    'momentum': 0.9,
    'weight_decay': 1e-4,

    'epochs': 200,
    'epoch_save': 1,
    'epoch_rcnn': 20,
    'num_workers': 30,

    'train_set_list': BASE + 'split_full_with_nodule_9classes/train.txt',
    'val_set_list': BASE + 'split_full_with_nodule_9classes/val.txt',
    'test_set_name': BASE + 'split_full_with_nodule_9classes/test.txt',
    'train_anno': BASE + 'split_full_with_nodule_9classes/train_anno.csv',
    'test_anno': BASE + 'data/luna16_annot_center.csv', # BASE + 'split_full_with_nodule_9classes/test_anno.csv',
    'DATA_DIR': data_config['preprocessed_data_dir'],
    'ROOT_DIR': os.getcwd()
}

if train_config['optimizer'] == 'SGD':
    train_config['init_lr'] = 0.01
elif train_config['optimizer'] == 'Adam':
    train_config['init_lr'] = 0.001
elif train_config['optimizer'] == 'RMSprop':
    train_config['init_lr'] = 2e-3


train_config['RESULTS_DIR'] = os.path.join(train_config['ROOT_DIR'], 'results')
train_config['out_dir'] = os.path.join(train_config['RESULTS_DIR'], 'full01')
train_config['initial_checkpoint'] = None #


config = dict(data_config, **net_config)
config = dict(config, **train_config)



import torch.utils.data as data
import data.util as util
import torchvision.transforms as transforms
import torchvision.transforms.functional as tf
from PIL import Image

class LRHRDataset(data.Dataset):
    '''
    Read LR and HR image pairs.
    If only HR image is provided, generate LR image on-the-fly.
    The pair is ensured by 'sorted' function, so please check the name convention.
    '''

    def __init__(self, opt):
        super(LRHRDataset, self).__init__()
        self.opt = opt
        self.paths_LR = None
        self.paths_HR = None
        self.LR_env = None  # environment for lmdb
        self.HR_env = None

        self.scale = self.opt['scale']
        self.HR_size = self.opt['HR_size']

        # if self.opt['phase'] == 'train':
        #     if  self.opt['use_rot']:
        #         self.transform = transforms.Compose([
        #                                              # transforms.ToPILImage(),
        #                                              # BILINEAR makes images smoother
        #                                              transforms.RandomRotation(20,resample=Image.BILINEAR),
        #                                              transforms.RandomCrop(self.HR_size),
        #                                              transforms.RandomResizedCrop(self.HR_size,
        #                                                                          scale=(0.8, 1.0),
        #                                                                          ratio=(1.0, 1.0)),
        #
        #
        #                                             ])
        #     else:
        #         self.transform = transforms.Compose([
        #                                              # transforms.ToPILImage(),
        #                                              transforms.RandomCrop(self.HR_size),
        #                                             ])
        # else: # inference mode
        #     self.transform = transforms.Compose([transforms.ToPILImage(),
        #                                          ])

        self.ToTensor = util.ImgToTensor()

        # read image list from lmdb or image files
        self.HR_env, self.paths_HR = util.get_image_paths(opt['data_type'], opt['dataroot_HR'])
        self.LR_env, self.paths_LR = util.get_image_paths(opt['data_type'], opt['dataroot_LR'])

        assert self.paths_HR, 'Error: HR path is empty.'
        if self.paths_LR and self.paths_HR:
            assert len(self.paths_LR) == len(self.paths_HR), \
                'HR and LR datasets have different number of images - {}, {}.'.format(\
                len(self.paths_LR), len(self.paths_HR))

    def _transform(self, input, target):

        if self.opt['use_rot']:# random rotation
            angle = transforms.RandomRotation.get_params([-20, 20])
            input = tf.rotate(input, angle, resample=Image.BILINEAR)
            target = tf.rotate(target, angle, resample=Image.BILINEAR)

        params = transforms.RandomCrop.get_params(input, (self.HR_size, self.HR_size))
        input = tf.crop(input, *params)
        target = tf.crop(target, *params)

        if self.opt['use_zoom']:# random zoom in
            params = transforms.RandomResizedCrop.get_params(input, scale=(0.7, 1.0), ratio=(1.0, 1.0))
            input = tf.resized_crop(input, *params, self.HR_size)
            target = tf.resized_crop(target, *params, self.HR_size)

        return input, target

    # def __getitem__(self, index):
    #
    #     HR_path, LR_path = None, None
    #
    #     # get HR image
    #     HR_path = self.paths_HR[index]
    #     img_HR = util.read_img(self.HR_env, HR_path)
    #     # convert pillow mode to F so we can use resize function, pillow does not support int16
    #     # shame on you pillow! Very inconvenient
    #
    #     img_HR = self.transform(transforms.ToPILImage()(img_HR).convert('F'))
    #     # img_HR = self.transform(img_HR).convert('F')
    #
    #     # get LR image on the fly
    #
    #     if self.paths_LR:
    #     # already have LR images, perform the same transform
    #         LR_path = self.paths_LR[index]
    #         img_LR = util.read_img(self.LR_env, LR_path)
    #         img_LR = self.transform(transforms.ToPILImage()(img_LR).convert('F'))
    #     else:
    #         width = self.HR_size if self.opt['phase'] == 'train' else 512
    #         img_LR = transforms.Resize(width // self.scale)(img_HR)
    #
    #     # concert everything to tensor
    #     img_LR = self.ToTensor(img_LR)
    #     img_HR = self.ToTensor(img_HR)
    #
    #     if LR_path is None:
    #         LR_path = HR_path
    #     return {'LR': img_LR, 'HR': img_HR, 'LR_path': LR_path, 'HR_path': HR_path}

    def __getitem__(self, index):
        LR_path, HR_path = None, None
        # get HR image
        HR_path = self.paths_HR[index]
        img_HR = util.read_img(self.HR_env, HR_path)
        # convert pillow mode to F so we can use resize function, pillow does not support int16
        img_HR = transforms.ToPILImage()(img_HR).convert('F')

        # get LR image on the fly
        if self.paths_LR:
        # already have LR images, convert to pil images
            LR_path = self.paths_LR[index]
            img_LR = util.read_img(self.LR_env, LR_path)
            img_LR = transforms.ToPILImage()(img_LR).convert('F')
            if self.opt['phase'] == 'train':
                img_LR, img_HR = self._transform(img_LR, img_HR)
        else:
            # mimic the degradation by downsample
            width = self.HR_size if self.opt['phase'] == 'train' else 512
            img_LR = transforms.Resize(width // self.scale)(img_HR)

        # convert everything to tensor
        img_LR, img_HR = self.ToTensor(img_LR), self.ToTensor(img_HR)

        if LR_path is None:
            LR_path = HR_path
        return {'LR': img_LR, 'HR': img_HR, 'LR_path': LR_path, 'HR_path': HR_path}

    def __len__(self):
        return len(self.paths_HR)



import os, glob, time
import logging
import argparse
from collections import OrderedDict
import numpy as np
import torch
import copy
import options.options as option
import utils.util as util
from data import create_dataset, create_dataloader
from models import create_model
import random


def main():
    # options
    parser = argparse.ArgumentParser()
    parser.add_argument('-opt', type=str, required=True, help='Path to options JSON file.')
    opt = option.parse(parser.parse_args().opt, is_train=False)
    util.mkdirs((path for key, path in opt['path'].items() if not key == 'pretrain_model_G'))
    opt = option.dict_to_nonedict(opt)
    # set up logger
    util.setup_logger(None, opt['path']['log'], 'test.log', level=logging.INFO, screen=True)
    logger = logging.getLogger('base')
    logger.info(option.dict2str(opt))
    torch.backends.cudnn.benchmark = True
    
    test_loaders = []
    need_label, data_merged = opt['need_label'], opt['data_merged']
    need_atten = opt['network_G']['use_attention']
    if need_atten:
        need_candidate_atten = opt['network_G']['use_attention']['generate_candidate']
    else:
        need_candidate_atten = opt['network_G']['use_attention']

    for phase, dataset_opt in opt['datasets'].items():
        dataset_opt['need_label'] = need_label
        dataset_opt['data_merged'] = data_merged
        dataset_opt['use_candidate_atten'] = need_candidate_atten
        test_set = create_dataset(dataset_opt)
        test_loader = create_dataloader(test_set, dataset_opt) 
        logger.info('Number of test volumes in [{:s}]: {:d}'.format(dataset_opt['name'], len(test_set)))
        test_loaders.append(test_loader)

    # data = iter(test_loaders[0]).next()
    # print("batch data keys:", data.keys())
    # print('LR shape:', data['LR'].shape)
    # print('HR shape:', data['HR'].shape)
    # print('Mask shape:', data['mask'].shape)

    # create model
    model = create_model(opt)
    # copy net and convert to fp16 if necessary
    model.half()
    # create pdist model vgg
    pdist_model = util.create_pdist_model(use_gpu=opt['gpu_ids'] is not None)
    if opt["precision"] == 'int8':
        model.prepare_quant(test_loaders[0])
    for test_loader in test_loaders:
        test_set_name = test_loader.dataset.opt['name']
        logger.info('\nTesting [{:s}]...'.format(test_set_name))
        results_dir = os.path.join(opt['path']['results_root'], test_set_name)
        util.mkdir(results_dir)
        pnsr_results = OrderedDict()
        ssim_results = OrderedDict()
        pdist_results = OrderedDict()
        total_t = 0.

        for i, data in enumerate(test_loader):
            need_HR = False if test_loader.dataset.opt['dataroot_HR'] is None else True
            has_mask = False if test_loader.dataset.opt['maskroot_HR'] is None else True
            logger.info('start inference...')
            model.feed_test_data(data, need_HR=need_HR)
            t0 = time.time()
            model.test(data)  # test
            if opt['gpu_ids']:
                torch.cuda.synchronize()
            t1 = time.time()
            t = t1 - t0
            total_t += t
            logger.info('inference time: {:.3f}'.format(t))
            # get cpu numpy from cuda tensor
            visuals = model.get_current_visuals(data, maskOn=has_mask, need_HR=need_HR)
            # save volume data
            patient_id = data['uid'][0]
            save_vol_name = 'norm'+'_'+patient_id.split('.')[0]
            vol_path = os.path.join(results_dir, save_vol_name)
            if len(data['spacings']) != 0:
                LR_spacings = [x.item() for x in data['spacings']]
            else:
                LR_spacings = []

            if 'nrrd' in opt['result_format']:
                logger.info('saving nnrd...')
                sr_vol = util.tensor2img(visuals['SR'], out_type=np.uint16)
                util.save_vol(opt, LR_spacings, vol_path + '.nrrd', sr_vol)
            if 'dicom' in opt['result_format']:
                logger.info('saving dicoms...') 
                sr_vol = util.tensor2img(visuals['SR'], out_type=np.int16, intercept = -1000) 
                util.save_dicoms(opt, LR_spacings, vol_path, sr_vol)
            if  opt['result_format']!='nrrd' and opt['result_format']!='dicom':
                raise NotImplementedError('supported output format: nrrd or dicom')

            # initialize dictionary with empty {}
            pnsr_results[patient_id] = {}
            ssim_results[patient_id] = {}
            pdist_results[patient_id] = {}

            # calculate image evaluation metrics
            if need_HR:
                def _calculate_metrics(sr_vol, gt_vol, view='xy', avg_voxels=False):
                    sum_psnr = 0.
                    sum_ssim = 0.
                    sum_pdist = 0.

                    # [D,H,W]
                    num_val = 0 # psnr could be inf at xz or yz (near edges), will not calculate
                    for i, vol in enumerate(zip(sr_vol, gt_vol)):
                        sr_img, gt_img = vol[0], vol[1] # shape: [823, 512, 512] of type numpy.ndarray
                        # print('i is {}, sr_img shape is {} and gt_img shape is {}'.format(i, sr_img.shape, gt_img.shape))
                        # range is assume to be [0,255] so  have to scale back from 1500 to 255 float64
                        crop_size = round(opt['scale'])
                        # shape [510, 510]
                        cropped_sr_img = sr_img[crop_size:-crop_size, crop_size:-crop_size]\
                                            .astype(np.float64) / 1500. * 255.
                        # shape [510, 510]
                        cropped_gt_img = gt_img[crop_size:-crop_size, crop_size:-crop_size]\
                                            .astype(np.float64) / 1500. * 255.
                        psnr = util.calculate_psnr(cropped_sr_img, cropped_gt_img)
                        ssim = util.calculate_ssim(cropped_sr_img, cropped_gt_img)
                        pdist = util.calculate_pdist(pdist_model, cropped_sr_img, cropped_gt_img)
                        if psnr != float('inf'):
                            num_val += 1
                            sum_psnr += psnr
                            sum_ssim += ssim
                            sum_pdist += pdist

                        logger.info('{:20s} - {:3d}- PSNR: {:.6f} dB; SSIM: {:.6f}; pdist: {:.6f}'\
                                        .format(patient_id, i+1, psnr, ssim, pdist))

                    pnsr_results[patient_id][view] = sum_psnr / num_val
                    ssim_results[patient_id][view] = sum_ssim / num_val
                    pdist_results[patient_id][view] = sum_pdist / num_val
                    return pnsr_results, ssim_results, pdist_results
                
                sr_vol = util.tensor2img(visuals['SR'], out_type=np.uint16) 
                gt_vol = util.tensor2img(visuals['HR'], out_type=np.uint16)  # uint16 range [0,1500]
                min_depth = min(sr_vol.shape[0], gt_vol.shape[0])
                sr_vol = sr_vol[:min_depth,...]
                gt_vol = gt_vol[:min_depth,...] # make sure they have the same depth
                
                # [H W] axial view
                _calculate_metrics(sr_vol, gt_vol, view='xy')
                # [D W] coronal view
                _calculate_metrics(sr_vol.transpose(1, 0, 2), gt_vol.transpose(1, 0, 2), view='xz')
                # [D H] sagittal view
                _calculate_metrics(sr_vol.transpose(2, 0, 1), gt_vol.transpose(2, 0, 1), view='yz')
            else:
                logger.info(patient_id)

        if need_HR:  # metrics
            # print result dictionary
            util.print_metrics(logger, 'test PSNR', pnsr_results)
            util.print_metrics(logger, 'test SSIM', ssim_results)
            util.print_metrics(logger, 'test pdist', pdist_results)
        logger.info('average inference time: {:.3f}'.format(total_t/len(test_loader)))


if __name__ == '__main__':
    main()


import os
import re
import random
import numpy as np
import json
import nrrd
from sklearn.model_selection import train_test_split


def cvt_int(inputString):
    find_digit = re.search(r'\d', inputString)
    reformat_num = int(float(inputString[find_digit.start():]))
    reformat_str = inputString[:find_digit.start()] + str(reformat_num)
    return reformat_str

def rename_file_with_cd(data_path):
    img_conditions = os.listdir(data_path)
    for each_img_cd in img_conditions:
        dir_path = os.path.join(data_path, each_img_cd)
        cd_info = '_'.join([cvt_int(cd) for cd in dir_path.split('/')[-1].split('_')])
        for file in os.listdir(dir_path):
            file_name, ext = file.split('.')
            new_name = file_name + '_' + cd_info + '.' + ext
            new_name_path = os.path.join(dir_path, new_name)
            os.rename(os.path.join(dir_path, file), new_name_path)
    print('FILES RENAMED!')

def write_txt_file(data_list, name, path):
    file_name = '{}.txt'.format(name)
    file_path = os.path.join(path, file_name)
    with open(file_path, 'w') as f:
        f.write('\n'.join(data_list))

def generate_specific_uids(path_to_data, path_to_save_txt, split, need_testSet_withNod):
    mapping_uids = {}
    train_uids, val_uids, test_uids = [], [], []
    for file in os.listdir(path_to_data):
        condition = '_'.join(file.split('.')[0].split('_')[1:])
        if condition not in mapping_uids:
            mapping_uids[condition] = []
        file_name = file.split('.')[0]
        if file_name not in mapping_uids[condition]:
            mapping_uids[condition].append(file_name)
    # random or based on ref split
    if split == 'random':
        # get uid for each condition
        for key in mapping_uids:
            uids = mapping_uids[key]
            random.shuffle(uids)
            train_tmp, test_uid = train_test_split(uids, test_size=0.15)
            train_uid, val_uid = train_test_split(train_tmp, test_size=0.15)
            train_uids.extend(train_uid)
            val_uids.extend(val_uid)
            test_uids.extend(test_uid)
            """
            train_uids.extend(random.sample(train_uid, 15))
            val_uids.extend(random.sample(val_uid, 3))
            test_uids.extend(random.sample(test_uid, 4))
            """
    else:
        read_ref = [file.split('.h5')[0] for file in os.listdir(split) if file.endswith('.h5')]
        ref_test_uid = []
        nodule_count = 0
        if need_testSet_withNod:
            with open(need_testSet_withNod, 'r') as file:
                ucla_nodule_annot = json.load(file)
            for annot in ucla_nodule_annot:
                if nodule_count > 22:
                    break
                nodule_count += len(ucla_nodule_annot[annot])
                ref_test_uid.append(annot)
            ref_test_uid = ref_test_uid[:15]
            ref_train_val = [uid for uid in read_ref if uid not in ref_test_uid]
            ref_train_uid, ref_val_uid = train_test_split(ref_train_val, test_size=0.15)
        else:
            ref_train_tmp, ref_test_uid = train_test_split(read_ref, test_size=0.15)
            ref_train_uid, ref_val_uid = train_test_split(ref_train_tmp, test_size=0.15)
        # append uids
        for key in mapping_uids:
            uids = [file.split('_')[0] for file in mapping_uids[key]]
            for uid in uids:
                if uid in ref_train_uid:
                    train_uids.append(uid+'_'+key)
                elif uid in ref_val_uid:
                    val_uids.append(uid+'_'+key)
                elif uid in ref_test_uid:
                    test_uids.append(uid+'_'+key)
                else:
                    print('Skiping UID:', uid)
    # shuffle merged lists and write to file
    random.shuffle(train_uids)
    random.shuffle(val_uids)
    random.shuffle(test_uids)      
    write_txt_file(train_uids, 'train_all_ep', path_to_save_txt)
    write_txt_file(val_uids, 'valid_all_ep', path_to_save_txt)
    write_txt_file(test_uids, 'test_all_ep', path_to_save_txt)
    print('UIDS GENERATED! Train: {}, Val: {}, Test: {}'.format(len(train_uids), len(val_uids), len(test_uids)))

def generate_nodVol_split(path_to_vol, path_to_save_txt, need_external_test):
    files = [file for file in os.listdir(path_to_vol) if file.endswith('.nrrd')]
    print('Total files:', len(files))
    random.shuffle(files)
    train_uids, val_uids, test_uids = [], [], []
    data_dict = {'nod':[], 'nonod':[]}
    for uid in files:
        nod_type = uid.split('_')[0]
        data_dict[nod_type].append(uid)
    print('Total nodules {} and non-nodules {}'.format(len(data_dict['nod']), len(data_dict['nonod'])))
    if need_external_test:
        with open(need_external_test, 'r') as f:
            lines = f.readlines()
            test_cases = [l.rstrip() for l in lines]
        unique_test = []
        for case in test_cases:
            _ = case.split('_')[0]
            if _ not in unique_test:
                unique_test.append(_)
        allowed_uniques = []
        for file in data_dict['nod']:
            _ = file.split('_')[1]
            if _ in unique_test:
                allowed_uniques.append(file)
        random.shuffle(allowed_uniques)
        test_uids.extend(random.sample(allowed_uniques, 22))
        test_uids.extend(random.sample(data_dict['nonod'], 22))
        # print(len(test_uids))
        new_data_dict = {'nod':list(set(data_dict['nod'])-set(test_uids)), 'nonod':list(set(data_dict['nonod'])-set(test_uids))}
        # print(len(new_data_dict['nod']), len(new_data_dict['nonod']))
        for data in new_data_dict:
            uids = new_data_dict[data]
            ref_train_uid, ref_val_uid = train_test_split(uids, test_size=0.15)
            train_uids.extend(ref_train_uid)
            val_uids.extend(ref_val_uid)
    else:
        for data in data_dict:
            uids = data_dict[data]
            ref_train_tmp, ref_test_uid = train_test_split(uids, test_size=0.15)
            ref_train_uid, ref_val_uid = train_test_split(ref_train_tmp, test_size=0.15)
            train_uids.extend(ref_train_uid)
            val_uids.extend(ref_val_uid)
            test_uids.extend(ref_test_uid)
    # |----- shuffle uids ------------|
    random.shuffle(train_uids)
    random.shuffle(val_uids)
    random.shuffle(test_uids)
    # |----- write to file ------------|
    write_txt_file(train_uids, 'train_nodule', path_to_save_txt)
    write_txt_file(val_uids, 'val_nodule', path_to_save_txt)
    write_txt_file(test_uids, 'test_nodule', path_to_save_txt)
    print('UIDs Created!')

def remove_files(path_to_cases, cases_to_keep):
    with open(cases_to_keep, 'r') as f:
        lines = f.readlines()
        test_set_uids = [l.rstrip() for l in lines]
        for file in os.listdir(path_to_cases):
            file_name = file.split('.nrrd')[0]
            if file_name not in test_set_uids:
                os.remove(os.path.join(path_to_cases, file))
        print('FILES REMOVED!')

def seperate_cases(file_1, file_2):
    uids_to_exclude, uids_to_keep, nodule_annot_to_keep = [], [], []
    with open(file_1, 'r') as f:
        lines = f.readlines()
        set_uids = [l.rstrip().split("_")[0] for l in lines if 'k2_d25' in l]
        uids_to_exclude.extend(set_uids)
    with open(file_2, 'r') as f:
        lines = f.readlines()
        set_uids = [l.rstrip().split("_")[0] for l in lines if 'k2_d25' in l]
        uids_to_exclude.extend(set_uids)
    with open('/datasets/nodule_annotations/ucla_nodule_uids.txt', 'r') as f:
        lines = f.readlines()
        set_uids = [l.rstrip() for l in lines]
        nodule_annot_to_keep.extend(set_uids)
    all_uids = '/datasets/data/k2_d25_st1.0'
    for file in os.listdir(all_uids):
        if file.endswith('.h5'):
            file_name = file.split('.')[0]
            if file_name not in uids_to_exclude and file_name in nodule_annot_to_keep:
                uids_to_keep.append(file_name)
    for uid in uids_to_keep:
        print(uid)

def check_size(uid_path, vol_path):
    with open(uid_path, 'r') as file:
        lines = file.readlines()
        lines = [line.strip() for line in lines]

    for line in lines:
        img_path = os.path.join(vol_path, line)
        data, header = nrrd.read(img_path)
        if data.shape[0] != 64 or data.shape[1] != 64 or data.shape[2] != 16:
            print(line)



if __name__ == '__main__':
    # path_to_data = '/datasets/data'
    # rename_file_with_cd(path_to_data)
    # path_to_merged_data = '/datasets/data_st1.0_merged'
    # generate_merged_uids(path_to_merged_data, '/datasets/uids/')
    # generate_specific_uids(path_to_merged_data, '/datasets/uids/merged_uids', split='/datasets/reference/k2_d100_st1.0',
    #                         need_testSet_withNod='/datasets/nodule_annotations/ucla_annotations.json')
    # remove_files('/workspace/cNormGAN_debug/results/SNGAN-Embed-Gen-Disc-allTileStitch-k1d10-k2d100-Nastaran/test_st1.0_nastaran/', 
    #                 '/datasets/nodule_annotations/ucla_nastaran_subset_32_test_uid.txt')
    # seperate_cases('/datasets/uids/merged_uids/merged_train_subset_uids.txt', '/datasets/uids/merged_uids/merged_valid_subset_uids.txt')
    # check_files('/datasets/uids/merged_uids/merged_train_subset_uids.txt', '/datasets/data_train_tmp', '/datasets/data_st1.0_merged')
    # path_to_nodVol = '/datasets/UCLA-Nas-NoduleVolume/unnormalized/k2_d100_st1.0'
    # generate_nodVol_split(path_to_nodVol, '/datasets/UCLA-Nas-NoduleVolume/uids', '/datasets/uids/merged_uids/merged_test_all_ep_uids.txt')
    check_size('/datasets/UCLA-Nas-NoduleVolume/uids/test_nodule.txt', '/datasets/UCLA-Nas-NoduleVolume/unnormalized/k2_d100_st1.0')




from collections import OrderedDict
import torch
import torch.nn as nn
from torch.nn.modules.utils import _triple


"""
Helper function, for selecting activation
- neg_slope: for leakyrelu and init of prelu
- n_prelu: for p_relu num_parameters
"""
def act(act_type, inplace=True, neg_slope=0.2, n_prelu=1):
    act_type = act_type.lower() # convert lowercase
    # intialize activation based on type
    if act_type == 'relu':
        layer = nn.ReLU(inplace)
    elif act_type == 'leakyrelu':
        layer = nn.LeakyReLU(neg_slope, inplace)
    elif act_type == 'prelu':
        layer = nn.PReLU(num_parameters=n_prelu, init=neg_slope)
    else:
        raise NotImplementedError('activation layer [{:s}] is not found'.format(act_type))
    return layer

"""
Helper function, for selecting normalization layer
"""
def norm(norm_type, nc):
    # helper selecting normalization layer
    norm_type = norm_type.lower()
    if norm_type == 'batch':
        layer = nn.BatchNorm3d(nc, affine=True)
    elif norm_type == 'instance':
        layer = nn.InstanceNorm3d(nc, affine=False)
    elif norm_type =='layer':
        layer = Layernorm(nc,affine=False)
    else:
        raise NotImplementedError('normalization layer [{:s}] is not found'.format(norm_type))
    return layer

"""
Note: If dilation is 1, it is a normal convolution operation. If the dilation rate is 2, then there is skip 
of one pixel per input. In general, if there is a dilation rate of n, skip of n-1 pixels per input. 
- Receptive field will increase as dilation rate is increased.
- Number of elements of filter remains the same but with the increase in dilation rate, they will cover 
more coverage.
If dilation is not 1, we calculate the padding size such that input image and output feature map is same
"""
def get_valid_padding(kernel_size, dilation):
    # print("checking for valid padding")
    # a = kernel_size
    kernel_size = kernel_size + (kernel_size - 1) * (dilation - 1)
    # print("kernel size = {} + ({} - 1) * ({} - 1) = {}".format(a, a, dilation, kernel_size))
    padding = (kernel_size - 1) // 2
    # print("padding = ({}-1/2) = {}".format(kernel_size, padding))
    return padding


"""
Helper function to create nn.Sequential module give a tuple data created from 'conv_block' data
"""
def sequential(*args):
    # print('sequential function called!')
    # raise error if input tuple has length 1 It unwraps nn.Sequential.
    if len(args) == 1:
        if isinstance(args[0], OrderedDict):
            raise NotImplementedError('sequential does not support OrderedDict input.')
        return args[0]  # No sequential is needed.
    modules = []
    for module in args:
        # if 'module' is instance of 'nn.Sequential', we unpack the childrens of 'module' since it will have
        # more than one layer.
        # if 'module' is instance of 'nn.Module', it will only have one layer so no need to unpack
        if isinstance(module, nn.Sequential):
            for submodule in module.children():
                modules.append(submodule)
        elif isinstance(module, nn.Module):
            modules.append(module)
    return nn.Sequential(*modules)

"""
Returns one colvolutions layer -> ConvLayer, Activation (if specified), BactchNorm (if specified)
"""
def conv_block(in_nc, out_nc, kernel_size, stride=1, dilation=1, bias=True, \
               norm_type=None, act_type='relu'):
    # get padding size, in case dilation is not 1, so that input image and feature map size is same
    padding = get_valid_padding(kernel_size, dilation) # padding = 1
    c = nn.Conv3d(in_nc, out_nc, kernel_size=kernel_size, stride=stride, padding=padding, \
            dilation=dilation, bias=bias)
    a = act(act_type) if act_type else None # get activation type if provided
    n = norm(norm_type, out_nc) if norm_type else None # get normalization type if provided
    return sequential(c, n, a)


def s_conv_block(in_nc, out_nc, kernel_size, stride=1, dilation=1, bias=True, \
               norm_type=None, act_type='relu'):
    '''
    Conv layer with padding, normalization, activation
    Conv -> Norm -> Activation
    '''
    padding = get_valid_padding(kernel_size, dilation)

    kernel_size = _triple(kernel_size)
    stride = _triple(stride)
    padding = _triple(padding)
        
    spatial_kernel_size =  [1, kernel_size[1], kernel_size[2]]
    spatial_stride =  [1, stride[1], stride[2]]
    spatial_padding =  [0, padding[1], padding[2]]

    c = nn.Conv3d(in_nc, out_nc, spatial_kernel_size, stride=spatial_stride, padding=spatial_padding, \
            dilation=dilation, bias=bias)
    a = act(act_type) if act_type else None
    n = norm(norm_type, out_nc) if norm_type else None

    return sequential(c, n, a)


def t_conv_block(in_nc, out_nc, kernel_size, stride=1, dilation=1, bias=True, \
               norm_type=None, act_type='relu'):
    '''
    Conv layer with padding, normalization, activation
    Conv -> Norm -> Activation
    '''
    padding = get_valid_padding(kernel_size, dilation)

    kernel_size = _triple(kernel_size)
    stride = _triple(stride)
    padding = _triple(padding)

    temporal_kernel_size = [kernel_size[0], 1, 1]
    temporal_stride =  [stride[0], 1, 1]
    temporal_padding =  [padding[0], 0, 0]

    c = nn.Conv3d(in_nc, out_nc, temporal_kernel_size, stride=temporal_stride, padding=temporal_padding, \
            dilation=dilation, bias=bias)
    a = act(act_type) if act_type else None
    n = norm(norm_type, out_nc) if norm_type else None

    return sequential(c, n, a)


class Layernorm(nn.Module):
    def __init__(self, num_features, eps=1e-5, affine=True):
        super(Layernorm, self).__init__()
        self.num_features = num_features
        self.affine = affine
        self.eps = eps

        if self.affine:
            self.gamma = nn.Parameter(torch.Tensor(num_features).uniform_())
            self.beta = nn.Parameter(torch.zeros(num_features))

    def forward(self, x):
        shape = [-1] + [1] * (x.dim() - 1)
        mean = x.view(x.size(0), -1).mean(1).view(*shape)
        std = x.view(x.size(0), -1).std(1).view(*shape)

        y = (x - mean) / (std + self.eps)
        if self.affine:
            shape = [1, -1] + [1] * (x.dim() - 2)
            y = self.gamma.view(*shape) * y + self.beta.view(*shape)
        return y


"""
Element-wise sum the output of a residual to its input
"""
class ShortcutBlock(nn.Module):
    # residual will be 'resnet_blocks' with 'lr_conv' combined together in nn.Sequential
    def __init__(self, residual):
        super(ShortcutBlock, self).__init__()
        self.res = residual
        self.skip_op = nn.quantized.FloatFunctional()

    def forward(self, x):
        return self.skip_op.add(x, self.res(x))

    # '__repr__' is a special method used to represent a class’s objects as a string 
    def __repr__(self):
        tmpstr = 'Identity + \n|'
        modstr = self.res.__repr__().replace('\n', '\n|')
        tmpstr = tmpstr + modstr
        return tmpstr




import SimpleITK as sitk
import torch.utils.data as data
import random
import os, glob
import json 
import torch
import numpy as np 
import nibabel as nib
import nrrd
import nibabel
import pydicom
from lung_analysis_pipeline.utils import util

class dcmDataset(data.Dataset):
    def __init__(self, opt): 
        super(dcmDataset, self).__init__()
        self.FILL_RATIO_THRESHOLD = 0.8
        self.opt = opt
        self.in_folder = opt['dataroot_LR']

        # feature extraction ROI 
        if opt['feature_roi_centroid']:
            with open(opt['feature_roi_centroid'], 'r') as json_file:
                self.feature_roi = json.load(json_file)
        if opt['feature_mask']:
            with open(opt['feature_mask'], 'r') as json_file: 
                self.img_to_mask_path = json.load(json_file)
        
        # 3d voxel size
        if self.opt['phase'] == 'train' or (self.opt['phase'] =='val' and self.opt['need_voxels']):
            self.ps = (opt['LR_slice_size'], opt['LR_size'], opt['LR_size'])

        self.paths, self.uids = self.get_instance_paths(opt['dataroot_LR'])

        self.scale = opt['scale']
        self.ToTensor = util.ImgToTensor()

    def __len__(self): 
        return len(self.paths)
    
    def __getitem__(self, index):
        ## load dicoms 
        # data_dir_path = os.path.join(self.paths[index], 'DICOM')
        # mask_dir_path = os.path.join(self.paths[index], 'Contours', 'NIFTI')
        # splits = os.path.normpath(data_dir_path).split(os.sep)
        # LR: extract volume cube during training, else return volume during validation 
        # dcm_files = [os.path.join(data_dir_path, f) for f in os.listdir(data_dir_path) if f.endswith('.dcm')]
        # dcm_files.sort(key=lambda x: int(pydicom.dcmread(x).InstanceNumber))
        # slices = [pydicom.dcmread(f) for f in dcm_files]
        # vol_in = np.stack([slice.pixel_array for slice in slices], axis=0).astype(np.float16)

        ## load NIFTI 
        data_dir_path = self.paths[index]
        data_file_path = os.path.join(data_dir_path, 'Contours', 'NIFTI', 'image.nii.gz')
        # mask_dir_path = os.path.join(data_dir_path, 'Contours', 'NIFTI')
        mask_dir_path = self.img_to_mask_path[data_dir_path]
        image_data = nib.load(data_file_path)
        vol_in = image_data.get_fdata()
        vol_in = np.transpose(vol_in, (2,1,0)) # onvert from WHD format to DHW format
        # Optional: Print the image shape and data type
        # print('Image Shape:', vol_in.shape)
        # print('Data Type:', vol_in.dtype)
        gt_mask = self.load_masks(mask_dir_path)
        # gt_mask = np.expand_dims(gt_mask, axis=0)
        # print('gt_mask Shape:', gt_mask.shape)
        # print('gt_mask Type:', gt_mask.dtype)

        # creates an extra dimension. e.g. (156, 512, 512) becomes (1, 156, 512, 512)
        vol_in = np.expand_dims(vol_in, axis=0)

        # convert to tensors and also within a certain range of values as defined in 'self.ToTensor()' class
        vol_in = self.ToTensor(vol_in, clip=True, raw_data_range=1500.)
        # vol_in = torch.from_numpy(np.array(vol_in, np.float32, copy=False))

        # generate a volume id 
        # patient_id = slices[0].PatientID.replace("_", "-")
        # timepoint = slices[0].StudyDate
        # series_instance_uid = slices[0].SeriesInstanceUID
        # modality = slices[0].Modality 
        # volume_id = f"{patient_id}_{timepoint}_{series_instance_uid}_{modality}"

        # # some info needed when saving results 
        # spacings = [slices[0].PixelSpacing[0], slices[0].PixelSpacing[1], abs(slices[1].SliceLocation - slices[0].SliceLocation)]
        # input_info = {'patient_id': splits[3], 'time': splits[4], 'name': splits[5]}
        # input_info['spacings'] = spacings
        # input_info['image_position'] = slices[0].ImagePositionPatient

        # load ROI centroids for feature extraction if provided 
        feature_rois = []
        if hasattr(self, 'feature_roi'): 
            feature_rois = self.feature_roi[data_dir_path]
            feature_rois = torch.tensor(feature_rois)
        # out_dict = {'LR': vol_in, 'spacings': [], 'uid': volume_id, 'feature_roi_centroids': feature_rois}
        out_dict = {'LR': vol_in, 'spacings': [], 'uid': self.uids[index], 'feature_roi_centroids': feature_rois, 'gt_mask': gt_mask}
        return out_dict
    
    def load_masks(self, mask_root): 
        combined_mask = None 
        for filename in os.listdir(mask_root): 
            if filename.startswith("Lesion"):
                file_path = os.path.join(mask_root, filename)
                mask = nib.load(file_path).get_fdata().astype(np.uint8)
                mask = np.transpose(mask, (2,1,0))
                if combined_mask is None:
                    combined_mask = mask
                else:
                    combined_mask = np.logical_or(combined_mask, mask).astype(np.uint8)

        return combined_mask
    """
    Returns a list of paths of instances. Assumes file structure to be the same as NLST_CT_annotation. 
    """
    def get_instance_paths(self, data_root): 
        paths = [] 
        ids = [] 
        for patient in os.listdir(data_root): 
            id = ''
            patient_path = os.path.join(data_root, patient)
            patient_id = patient.replace("_", "-") + '_'
            if not os.path.isdir(patient_path): 
                continue 
            for timepoint in os.listdir(patient_path): 
                timepoint_id = patient_id + "".join(timepoint.split("-")) + '_'
                time_path = os.path.join(patient_path, timepoint)
                for f in os.listdir(time_path): 
                    instance_id = timepoint_id + f.split('-')[-1] + '_CT'
                    paths.append( os.path.join(time_path, f) )
                    ids.append(instance_id) # [-1]
        return paths, ids


import torch.nn as nn
import torch.nn.functional as F
from . import block as B
import torch.nn.utils as utils
import torch


"""
Label Concatenation Conditioning with Auxially Loss on VGG style Discriminator with Spectral Normalization
"""
class Aux_Conditioned_Discriminator_VGG_64_SN(nn.Module):
    def __init__(self, in_nc, base_nf, kernel_class, dose_class):
        super(Aux_Conditioned_Discriminator_VGG_64_SN, self).__init__()
        self.disc_kernel_embedding = nn.Sequential(nn.Embedding(kernel_class, kernel_class), nn.Linear(kernel_class, 32*64*64))
        self.disc_dose_embedding = nn.Sequential(nn.Embedding(dose_class, dose_class), nn.Linear(dose_class, 32*64*64))
        self.lrelu = nn.LeakyReLU(0.2, True)
        """
        Apply spectral normalization to a paramaters in a given module.
        - Spectral normalization stabilizes the training of discriminators (critics) 
        in Generative Adversarial Networks (GANs) by rescaling the weight tensor with spectral
        norm \sigmaσ of the weight matrix calculated using power iteration method.
        """
        self.conv0 = utils.spectral_norm(nn.Conv3d(in_nc+2, base_nf, kernel_size=3, stride=1, padding=1))
        self.conv1 = utils.spectral_norm(nn.Conv3d(base_nf, base_nf, kernel_size=3, stride=(1, 2, 2), padding=1))
        # 32, 64
        self.conv2 = utils.spectral_norm(nn.Conv3d(base_nf, base_nf*2, kernel_size=3, stride=1, padding=1))
        self.conv3 = utils.spectral_norm(nn.Conv3d(base_nf*2, base_nf*2, kernel_size=3, stride=2, padding=1))
        # 16, 128
        self.conv4 = utils.spectral_norm(nn.Conv3d(base_nf*2, base_nf*4, kernel_size=3, stride=1, padding=1))
        self.conv5 = utils.spectral_norm(nn.Conv3d(base_nf*4, base_nf*4, kernel_size=3, stride=2, padding=1))
        # 8, 256
        self.conv6 = utils.spectral_norm(nn.Conv3d(base_nf*4, base_nf*8, kernel_size=3, stride=1, padding=1))
        self.conv7 = utils.spectral_norm(nn.Conv3d(base_nf*8, base_nf*8, kernel_size=3, stride=2, padding=1))
        # 4, 512
        self.linear0 = utils.spectral_norm(nn.Linear(512 * 4 * 4 * 4, 512))
        # classifier - real/fake image
        self.linear1 = utils.spectral_norm(nn.Linear(512, 1))
        # aux-classifier fc
        self.fc_aux_kernel = nn.Linear(512, kernel_class)
        self.fc_aux_dose = nn.Linear(512, dose_class)

    def forward(self, x, disc_kernel_lbl, disc_dose_lbl):   # input size is 64x64
        disc_kernel_embed = self.disc_kernel_embedding(disc_kernel_lbl).view(-1, x.shape[1], x.shape[2], x.shape[3], x.shape[4])
        disc_dose_embed = self.disc_dose_embedding(disc_dose_lbl).view(-1, x.shape[1], x.shape[2], x.shape[3], x.shape[4])
        x = torch.cat((x, disc_kernel_embed, disc_dose_embed), dim=1)
        x = self.lrelu(self.conv0(x))
        x = self.lrelu(self.conv1(x))
        x = self.lrelu(self.conv2(x))
        x = self.lrelu(self.conv3(x))
        x = self.lrelu(self.conv4(x))
        x = self.lrelu(self.conv5(x))
        x = self.lrelu(self.conv6(x))
        x = self.lrelu(self.conv7(x))
        x = x.view(x.size(0), -1)
        x = self.lrelu(self.linear0(x))
        # generate 3 branch predictions
        realfake = self.linear1(x)
        fc_aux_kernel = F.log_softmax(self.fc_aux_kernel(x), dim=1)
        fc_aux_dose = F.log_softmax(self.fc_aux_dose(x), dim=1)
        return realfake, fc_aux_kernel, fc_aux_dose

"""
Label Concatenation Conditioning on VGG style Discriminator with Spectral Normalization
"""
class Conditioned_Discriminator_VGG_64_SN(nn.Module):
    def __init__(self, in_nc, base_nf, kernel_class, dose_class):
        super(Conditioned_Discriminator_VGG_64_SN, self).__init__()
        self.disc_kernel_embedding = nn.Sequential(nn.Embedding(kernel_class, kernel_class), nn.Linear(kernel_class, 32*64*64))
        self.disc_dose_embedding = nn.Sequential(nn.Embedding(dose_class, dose_class), nn.Linear(dose_class, 32*64*64))
        self.lrelu = nn.LeakyReLU(0.2, True)
        """
        Apply spectral normalization to a paramaters in a given module.
        - Spectral normalization stabilizes the training of discriminators (critics) 
        in Generative Adversarial Networks (GANs) by rescaling the weight tensor with spectral
        norm \sigmaσ of the weight matrix calculated using power iteration method.
        """
        self.conv0 = utils.spectral_norm(nn.Conv3d(in_nc+2, base_nf, kernel_size=3, stride=1, padding=1))
        self.conv1 = utils.spectral_norm(nn.Conv3d(base_nf, base_nf, kernel_size=3, stride=(1, 2, 2), padding=1))
        # 32, 64
        self.conv2 = utils.spectral_norm(nn.Conv3d(base_nf, base_nf*2, kernel_size=3, stride=1, padding=1))
        self.conv3 = utils.spectral_norm(nn.Conv3d(base_nf*2, base_nf*2, kernel_size=3, stride=2, padding=1))
        # 16, 128
        self.conv4 = utils.spectral_norm(nn.Conv3d(base_nf*2, base_nf*4, kernel_size=3, stride=1, padding=1))
        self.conv5 = utils.spectral_norm(nn.Conv3d(base_nf*4, base_nf*4, kernel_size=3, stride=2, padding=1))
        # 8, 256
        self.conv6 = utils.spectral_norm(nn.Conv3d(base_nf*4, base_nf*8, kernel_size=3, stride=1, padding=1))
        self.conv7 = utils.spectral_norm(nn.Conv3d(base_nf*8, base_nf*8, kernel_size=3, stride=2, padding=1))
        # 4, 512

        # classifier
        self.linear0 = utils.spectral_norm(nn.Linear(512 * 4 * 4 * 4, 512))
        self.linear1 = utils.spectral_norm(nn.Linear(512, 1))

    def forward(self, x, disc_kernel_lbl, disc_dose_lbl):   # input size is 64x64
        disc_kernel_embed = self.disc_kernel_embedding(disc_kernel_lbl).view(-1, x.shape[1], x.shape[2], x.shape[3], x.shape[4])
        disc_dose_embed = self.disc_dose_embedding(disc_dose_lbl).view(-1, x.shape[1], x.shape[2], x.shape[3], x.shape[4])
        x = torch.cat((x, disc_kernel_embed, disc_dose_embed), dim=1)
        x = self.lrelu(self.conv0(x))
        x = self.lrelu(self.conv1(x))
        x = self.lrelu(self.conv2(x))
        x = self.lrelu(self.conv3(x))
        x = self.lrelu(self.conv4(x))
        x = self.lrelu(self.conv5(x))
        x = self.lrelu(self.conv6(x))
        x = self.lrelu(self.conv7(x))
        x = x.view(x.size(0), -1)
        x = self.lrelu(self.linear0(x))
        x = self.linear1(x)
        return x


"""
Auxillay Conditioning VGG style Discriminator with Spectral Normalization
"""
class Aux_Discriminator_VGG_64_SN(nn.Module):
    def __init__(self, in_nc, base_nf, kernel_class, dose_class):
        super(Aux_Discriminator_VGG_64_SN, self).__init__()
        self.lrelu = nn.LeakyReLU(0.2, True)
        """
        Apply spectral normalization to a paramaters in a given module.
        - Spectral normalization stabilizes the training of discriminators (critics) 
        in Generative Adversarial Networks (GANs) by rescaling the weight tensor with spectral
        norm \sigmaσ of the weight matrix calculated using power iteration method.
        """
        self.conv0 = utils.spectral_norm(nn.Conv3d(in_nc, base_nf, kernel_size=3, stride=1, padding=1))
        self.conv1 = utils.spectral_norm(nn.Conv3d(base_nf, base_nf, kernel_size=3, stride=(1, 2, 2), padding=1))
        # 32, 64
        self.conv2 = utils.spectral_norm(nn.Conv3d(base_nf, base_nf*2, kernel_size=3, stride=1, padding=1))
        self.conv3 = utils.spectral_norm(nn.Conv3d(base_nf*2, base_nf*2, kernel_size=3, stride=2, padding=1))
        # 16, 128
        self.conv4 = utils.spectral_norm(nn.Conv3d(base_nf*2, base_nf*4, kernel_size=3, stride=1, padding=1))
        self.conv5 = utils.spectral_norm(nn.Conv3d(base_nf*4, base_nf*4, kernel_size=3, stride=2, padding=1))
        # 8, 256
        self.conv6 = utils.spectral_norm(nn.Conv3d(base_nf*4, base_nf*8, kernel_size=3, stride=1, padding=1))
        self.conv7 = utils.spectral_norm(nn.Conv3d(base_nf*8, base_nf*8, kernel_size=3, stride=2, padding=1))
        # 4, 512
        self.linear0 = utils.spectral_norm(nn.Linear(512 * 4 * 4 * 4, 512))
        # classifier - real/fake image
        self.linear1 = utils.spectral_norm(nn.Linear(512, 1))
        # aux-classifier fc
        self.fc_aux_kernel = nn.Linear(512, kernel_class)
        self.fc_aux_dose = nn.Linear(512, dose_class)

    def forward(self, x):   # input size is 64x64
        x = self.lrelu(self.conv0(x))
        x = self.lrelu(self.conv1(x))
        x = self.lrelu(self.conv2(x))
        x = self.lrelu(self.conv3(x))
        x = self.lrelu(self.conv4(x))
        x = self.lrelu(self.conv5(x))
        x = self.lrelu(self.conv6(x))
        x = self.lrelu(self.conv7(x))
        x = x.view(x.size(0), -1)
        x = self.lrelu(self.linear0(x))
        # generate 3 branch predictions
        realfake = self.linear1(x)
        fc_aux_kernel = F.log_softmax(self.fc_aux_kernel(x), dim=1)
        fc_aux_dose = F.log_softmax(self.fc_aux_dose(x), dim=1)
        return realfake, fc_aux_kernel, fc_aux_dose


"""
Standard VGG style Discriminator with Spectral Normalization
"""
class Discriminator_VGG_64_SN(nn.Module):
    def __init__(self, in_nc, base_nf):
        super(Discriminator_VGG_64_SN, self).__init__()
        self.lrelu = nn.LeakyReLU(0.2, True)
        """
        Apply spectral normalization to a pramaters in a given module.
        - Spectral normalization stabilizes the training of discriminators (critics) 
        in Generative Adversarial Networks (GANs) by rescaling the weight tensor with spectral
        norm \sigmaσ of the weight matrix calculated using power iteration method.
        """
        self.conv0 = utils.spectral_norm(nn.Conv3d(in_nc, base_nf, kernel_size=3, stride=1, padding=1))
        self.conv1 = utils.spectral_norm(nn.Conv3d(base_nf, base_nf, kernel_size=3, stride=(1, 2, 2), padding=1))
        # 32, 64
        self.conv2 = utils.spectral_norm(nn.Conv3d(base_nf, base_nf*2, kernel_size=3, stride=1, padding=1))
        self.conv3 = utils.spectral_norm(nn.Conv3d(base_nf*2, base_nf*2, kernel_size=3, stride=2, padding=1))
        # 16, 128
        self.conv4 = utils.spectral_norm(nn.Conv3d(base_nf*2, base_nf*4, kernel_size=3, stride=1, padding=1))
        self.conv5 = utils.spectral_norm(nn.Conv3d(base_nf*4, base_nf*4, kernel_size=3, stride=2, padding=1))
        # 8, 256
        self.conv6 = utils.spectral_norm(nn.Conv3d(base_nf*4, base_nf*8, kernel_size=3, stride=1, padding=1))
        self.conv7 = utils.spectral_norm(nn.Conv3d(base_nf*8, base_nf*8, kernel_size=3, stride=2, padding=1))
        # 4, 512
        # classifier
        self.linear0 = utils.spectral_norm(nn.Linear(512 * 4 * 4 * 4, 512))
        self.linear1 = utils.spectral_norm(nn.Linear(512, 1))

    def forward(self, x):   # input size is 64x64
        x = self.lrelu(self.conv0(x))
        x = self.lrelu(self.conv1(x))
        x = self.lrelu(self.conv2(x))
        x = self.lrelu(self.conv3(x))
        x = self.lrelu(self.conv4(x))
        x = self.lrelu(self.conv5(x))
        x = self.lrelu(self.conv6(x))
        x = self.lrelu(self.conv7(x))
        x = x.view(x.size(0), -1)
        x = self.lrelu(self.linear0(x))
        x = self.linear1(x)
        return x


# VGG style Discriminator with input size 64*64
class Discriminator_VGG_64(nn.Module):
    def __init__(self, in_nc, base_nf, norm_type='batch', act_type='leakyrelu'):
        super(Discriminator_VGG_64, self).__init__()
        # features
        # batch x channel x t  x  h  x  w
        # batch x    1    x 32 x  64 x  64
        conv0 = B.conv_block(in_nc, base_nf, kernel_size=3, norm_type=None, act_type=act_type)
        conv1 = B.conv_block(base_nf, base_nf, kernel_size=3, stride=(1, 2, 2), norm_type=norm_type, act_type=act_type)
        # batch x    64   x 32 x  32 x  32
        conv2 = B.conv_block(base_nf, base_nf*2, kernel_size=3, stride=1, norm_type=norm_type, act_type=act_type)
        conv3 = B.conv_block(base_nf*2, base_nf*2, kernel_size=3, stride=2, norm_type=norm_type, act_type=act_type)
        # batch x    128  x 16 x  16 x  16
        conv4 = B.conv_block(base_nf*2, base_nf*4, kernel_size=3, stride=1, norm_type=norm_type, act_type=act_type)
        conv5 = B.conv_block(base_nf*4, base_nf*4, kernel_size=3, stride=2, norm_type=norm_type, act_type=act_type)
        # batch x    256  x 8 x   8 x  8
        conv6 = B.conv_block(base_nf*4, base_nf*8, kernel_size=3, stride=1, norm_type=norm_type, act_type=act_type)
        conv7 = B.conv_block(base_nf*8, base_nf*8, kernel_size=3, stride=2, norm_type=norm_type, act_type=act_type)
        # batch x    512  x 4 x   4 x  4

        # self.features = B.sequential(conv0, conv1, conv2, conv3, conv4, conv5 )
        self.features = B.sequential(conv0, conv1, conv2, conv3, conv4, conv5, conv6, conv7)
        # classifier
        self.classifier = nn.Sequential(
            # nn.Linear(256 * 8 * 8 * 8, 1024), nn.LeakyReLU(0.2, True), nn.Linear(1024, 1))
            nn.Linear(512* 4 * 4 * 4, 512), nn.LeakyReLU(0.2, True), nn.Linear(512, 1))

    def forward(self, x):
        x = self.features(x)
        x = x.view(x.size(0), -1)
        x = self.classifier(x)
        return x

# WGAN paper  https://arxiv.org/abs/1708.00961
# VGG style Discriminator with input size 64*64
class WGAN_Discriminator_VGG_64(nn.Module):
    def __init__(self, in_nc, base_nf, norm_type='batch', act_type='leakyrelu'):
        super(WGAN_Discriminator_VGG_64, self).__init__()
        # features
        # batch x channel x t  x  h  x  w [batch x    1    x 32 x  64 x  64]
        conv0 = B.conv_block(in_nc, base_nf, kernel_size=3, norm_type=None, act_type=act_type)
        conv1 = B.conv_block(base_nf, base_nf, kernel_size=3, stride=(1, 2, 2), norm_type=norm_type, act_type=act_type)
        # batch x    64   x 32 x  32 x  32
        conv2 = B.conv_block(base_nf, base_nf*2, kernel_size=3, stride=1, norm_type=norm_type, act_type=act_type)
        conv3 = B.conv_block(base_nf*2, base_nf*2, kernel_size=3, stride=2, norm_type=norm_type, act_type=act_type)
        # batch x    128  x 16 x  16 x  16
        conv4 = B.conv_block(base_nf*2, base_nf*4, kernel_size=3, stride=1, norm_type=norm_type, act_type=act_type)
        conv5 = B.conv_block(base_nf*4, base_nf*4, kernel_size=3, stride=2, norm_type=norm_type, act_type=act_type)
        # batch x    256  x 8 x   8 x  8

        self.features = B.sequential(conv0, conv1, conv2, conv3, conv4, conv5 )
        # classifier
        self.classifier = nn.Sequential(
            nn.Linear(256 * 8 * 8 * 8, 1024), nn.LeakyReLU(0.2, True), nn.Linear(1024, 1))

    def forward(self, x):
        x = self.features(x)
        x = x.view(x.size(0), -1)
        x = self.classifier(x)
        return x


"""
ESRGAN Discriminator
"""
class ESRGAN_Discriminator(nn.Module):
    def __init__(self, in_nc, base_nf):
        super(ESRGAN_Discriminator, self).__init__()
        self.features = nn.Sequential(
            nn.Conv3d(in_nc, base_nf, (3, 3, 3), (1, 1, 1), (1, 1, 1), bias=True),
            nn.LeakyReLU(0.2, True),
            nn.Conv3d(base_nf, base_nf, (4, 4, 4), (2, 2, 2), (1, 1, 1), bias=False),
            nn.BatchNorm3d(base_nf),
            nn.LeakyReLU(0.2, True),
            nn.Conv3d(base_nf, 128, (3, 3, 3), (1, 1, 1), (1, 1, 1), bias=False),
            nn.BatchNorm3d(128),
            nn.LeakyReLU(0.2, True),
            nn.Conv3d(128, 128, (4, 4, 4), (2, 2, 2), (1, 1, 1), bias=False),
            nn.BatchNorm3d(128),
            nn.LeakyReLU(0.2, True),
            nn.Conv3d(128, 256, (3, 3, 3), (1, 1, 1), (1, 1, 1), bias=False),
            nn.BatchNorm3d(256),
            nn.LeakyReLU(0.2, True),
            nn.Conv3d(256, 256, (4, 4, 4), (2, 2, 2), (1, 1, 1), bias=False),
            nn.BatchNorm3d(256),
            nn.LeakyReLU(0.2, True),
            nn.Conv3d(256, 512, (3, 3, 3), (1, 1, 1), (1, 1, 1), bias=False),
            nn.BatchNorm3d(512),
            nn.LeakyReLU(0.2, True),
            nn.Conv3d(512, 512, (4, 4, 4), (2, 2, 2), (1, 1, 1), bias=False),
            nn.BatchNorm3d(512),
            nn.LeakyReLU(0.2, True),
            nn.Conv3d(512, 512, (3, 3, 3), (1, 1, 1), (1, 1, 1), bias=False),
            nn.BatchNorm3d(512),
            nn.LeakyReLU(0.2, True),
            nn.Conv3d(512, 512, (4, 4, 4), (2, 2, 2), (1, 1, 1), bias=False),
            nn.BatchNorm3d(512),
            nn.LeakyReLU(0.2, True)
        )

        self.classifier = nn.Sequential(
            nn.Linear(512 * 2 * 2, 100),
            nn.LeakyReLU(0.2, True),
            nn.Linear(100, 1)
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # print('Inside Discriminator: incoming x shape:', x.shape)
        out = self.features(x)
        # print('out after features shape:', out.shape)
        out = torch.flatten(out, 1)
        # print('after flatten:', out.shape)
        out = self.classifier(out)
        return out


import os
import glob
import torch
import torch.nn as nn
import math
import lung_analysis_pipeline.Normalization_Pipeline.codes.models.networks as networks
import copy
import logging
logger = logging.getLogger('base')

class BaseModel():
    def __init__(self, opt):
        self.opt = opt
        self.device = torch.device('cuda' if opt['gpu_ids'] is not None else 'cpu') # gpu id is not defined
        self.is_train = opt['is_train'] # True for training
        self.schedulers = [] 
        self.optimizers = [] 
        # define network
        self.netG = networks.define_G(opt).to(self.device)
        

    """
    Creates class attributes for 'LR' and 'HR' data
    """
    def feed_train_data(self, data, need_HR=True):
        if self.opt['is_train'] and (self.opt['model'] == 'srgan' or self.opt['model'] == 'srgan_exp'):
            # get CT parameter label and candidates for attention
            if self.opt['network_G']['need_embed'] or self.opt['network_D']['need_embed'] or self.opt['network_D']['aux_lbl_loss']:
                if set(['kernel', 'dose']).issubset(data.keys()):
                    self.train_kernel_label = data['kernel'].to(self.device)
                    self.train_dose_label = data['dose'].to(self.device)
                else:
                    raise ValueError('Labels not found in `feed_train_data`')
            if self.opt['network_G']['use_attention']:
                if self.opt['network_G']['use_attention']['generate_candidate']:
                    if set(['Cand']).issubset(data.keys()):
                        self.cand_L_train = data['Cand'].to(self.device, non_blocking=True)
                    else:
                        raise ValueError('Candidates not found in `feed_train_data`')   
        # get LR and HR data
        self.var_L = data['LR'].to(self.device, non_blocking=True)  # LR
        if need_HR:
            self.real_H = data['HR'].to(self.device, non_blocking=True)  # HR


    """
    prepare LR and HR output to feed to model
    """
    def feed_test_data(self, data, need_HR=True):
        if self.opt['is_train']:
            if self.opt['model'] == 'srgan' or self.opt['model'] == 'srgan_exp':
                # get CT label and candidates for attention 
                if self.opt['network_G']['need_embed'] or self.opt['network_D']['need_embed'] or self.opt['network_D']['aux_lbl_loss']:
                    if set(['kernel', 'dose']).issubset(data.keys()):
                        self.test_kernel_label = data['kernel'].to(self.device)
                        self.test_dose_label = data['dose'].to(self.device)
                    else:
                        raise ValueError('Labels not found in `feed_test_data`')
                if self.opt['network_G']['use_attention']:
                    if self.opt['network_G']['use_attention']['generate_candidate']:
                        if set(['Cand']).issubset(data.keys()):
                            self.cand_L_test = data['Cand'].to(self.device, non_blocking=True)
                        else:
                            raise ValueError('Candidates not found in `feed_test_data`')
        elif not self.opt['is_train']:
            if self.opt['model'] == 'srgan' or self.opt['model'] == 'srgan_exp':
                if self.opt['network_G']['need_embed']:
                    if set(['kernel', 'dose']).issubset(data.keys()):
                        self.test_kernel_label = data['kernel'].to(self.device)
                        self.test_dose_label = data['dose'].to(self.device)
                    else:
                        raise ValueError('Labels not found in `feed_test_data`')
                    if self.opt['network_G']['use_attention']:
                        if self.opt['network_G']['use_attention']['generate_candidate']:
                            if set(['Cand']).issubset(data.keys()):
                                self.cand_L_test = data['Cand'].to(self.device, non_blocking=True)
                            else:
                                raise ValueError('Candidates not found in `feed_test_data`')
        # Get the LR and HR image
        if need_HR:
            self.real_H = data['HR'].to(self.device, non_blocking=True)
        opt_val = self.opt['datasets']['val']
        pt = opt_val['slice_size'] # 32
        self.ot = opt_val['overlap_slice_size'] # 4
        self.nt = 1 + math.ceil((data['LR'].size(2) - pt) / (pt - self.ot))
        # chunk 'slice_size' in z-direction if voxel is not needed
        if not opt_val['need_voxels']:
            # reshape the whole volume into blocks of voxel of size pt x 512 x 512
            self.var_L = torch.empty(self.nt, 1, pt, data['LR'].size(3), data['LR'].size(4)).to(self.device, non_blocking=True)
            # fill the volume
            for i in range(0, self.nt - 1):
                self.var_L[i, :, :, :, :] = data['LR'][0, 0, i*(pt-self.ot):i*(pt-self.ot)+pt, :, :]
            # the last one
            self.var_L[-1, :, :, :, :] = data['LR'][0, 0, -pt:, :, :]
        else:
            self.var_L = data['LR'].to(self.device, non_blocking=True)


    """
    Compute the current classification accuracy for auxially classes
    """
    def compute_acc(self, preds, labels):
        correct = 0
        preds_ = preds.data.max(1)[1]
        correct = preds_.eq(labels.data).cpu().sum()
        acc = float(correct) / float(len(labels.data)) * 100.0
        return acc

    
    """
    Copies fp32 model and convert to fp16
    """
    def half(self):
        logger.info("Model being converted to fp16")
        if self.opt['precision'] == 'fp16':
            self.netG_eval = copy.deepcopy(self.netG).half()
        else:
            self.netG_eval = self.netG
            
    def prepare_quant(self, loader, not_loader=False):
        # PyTorch Static Quantization 
        # https://leimao.github.io/blog/PyTorch-Static-Quantization/
        fused_model = copy.deepcopy(self.netG)
        fused_model.eval()
        # fused conv3d + relu
        fused_model = torch.quantization.fuse_modules(fused_model.net, [["3", "4"],["5","6"]], inplace=True)
        for i in range(8):
            torch.quantization.fuse_modules(fused_model[1].res[i].res, [["0", "1"]], inplace=True)
        quantized_model = networks.define_Quant(model_fp32=fused_model)
        # config
        quantization_config = torch.quantization.get_default_qconfig("fbgemm")
        quantized_model.qconfig = quantization_config
        torch.quantization.prepare(quantized_model, inplace=True)
        # calibration
        if not_loader:
            data = loader
        else:
            data = next(iter(loader))  
        # get a small chunk for calibration
        _ = quantized_model(data['LR'][:,:,:32 ,:,:])
        quantized_model = torch.quantization.convert(quantized_model, inplace=True)
        self.netG_eval = quantized_model

    def optimize_parameters(self):
        pass

    def get_current_visuals(self):
        pass

    def get_current_losses(self):
        pass

    def print_network(self):
        print("print_network called")
        pass

    def save(self, label):
        pass

    def load(self):
        print("load function called")
        pass


    """
    For each schedulers, we step up the learning rate
    """
    def update_learning_rate(self):
        for scheduler in self.schedulers:
            scheduler.step()


    """
    Returns the current learning rate of scheduler for generator only
    """
    def get_current_learning_rate(self):
        return self.schedulers[0].get_lr()[0]


    """
    Get string representation of a network and its parameters
    """
    def get_network_description(self, network):
        # check if 'network' is an instance of 'nn.DataParallel'
        if isinstance(network, nn.DataParallel):
            network = network.module
        s = str(network)
        n = sum(map(lambda x: x.numel(), network.parameters()))
        return s, n


    """
    Save the model (both generator & discriminator) 'state_dict' given the iteration step
    """
    def save_network(self, network, network_label, iter_step):
        # build weight name
        save_filename = '{}_{}.pth'.format(iter_step, network_label)
        # build path, with filename, to save weight
        save_path = os.path.join(self.opt['path']['models'], save_filename)
        # check if incoming 'network' is of 'DataParallel' instance
        if isinstance(network, nn.DataParallel):
            network = network.module
        # get state dict of network
        state_dict = network.state_dict()
        # put 'state_dict' parameters on cpu before saving, so the inference could even be done on cpu
        for key, param in state_dict.items():
            state_dict[key] = param.cpu()
        """
        # remove the previous files
        # old_files = glob.glob(os.path.join(self.opt['path']['models'], "*"+network_label+"*"))
        # for f in old_files:
        #     os.remove(os.path.join(self.opt['path']['models'], f))
        # save the new file
        """
        # save the models 
        torch.save(state_dict, save_path)


    """
    Loads the model with weights, given 'load_path' (generator or discriminator weight path), and
    the network architecture itself
    """
    def load_network(self, load_path, network, strict=True):
        # check if 'network' is an instance of 'DataParallel'
        if isinstance(network, nn.DataParallel):
            network = network.module
        network.load_state_dict(torch.load(load_path), strict=strict)


    """
    Saves training state (epoch, current_step, schedulers, optimizers) during training, which will be 
    used for resuming
    """
    def save_training_state(self, epoch, iter_step):
        # build state dictionary with information to save
        state = {'epoch': epoch, 'iter': iter_step, 'schedulers': [], 'optimizers': []}
        # save each schedulers 'state_dict()' to above dictionary
        for s in self.schedulers:
            state['schedulers'].append(s.state_dict())
        # save each optimizers 'state_dict()' to above dictionary
        for o in self.optimizers:
            state['optimizers'].append(o.state_dict())
        # build dictionary file name
        save_filename = '{}.state'.format(iter_step)
        # bulid path, with filename, to save dictionary
        save_path = os.path.join(self.opt['path']['training_state'], save_filename)
        
        # get all previous training state dictionary, if any
        old_files = os.listdir(self.opt['path']['training_state'])
        # remove the previous dictionary files, if there are multiple
        for f in old_files:
            os.remove(os.path.join(self.opt['path']['training_state'], f))
        # save the new dictionary
        torch.save(state, save_path)


    def resume_training(self, resume_state):
        '''Resume the optimizers and schedulers for training'''
        resume_optimizers = resume_state['optimizers']
        resume_schedulers = resume_state['schedulers']
        assert len(resume_optimizers) == len(self.optimizers), 'Wrong lengths of optimizers'
        assert len(resume_schedulers) == len(self.schedulers), 'Wrong lengths of schedulers'
        for i, o in enumerate(resume_optimizers):
            self.optimizers[i].load_state_dict(o)
        for i, s in enumerate(resume_schedulers):
            self.schedulers[i].load_state_dict(s)



import numpy as np
import h5py
import torch.utils.data as data
import lung_analysis_pipeline.utils.util as util
import random
import os
import re
import torch


class h5Dataset(data.Dataset):
    def __init__(self, dataset_opt):
        super(h5Dataset, self).__init__()
        # useful when extracting 3D mask cube from the input masks
        self.FILL_RATIO_THRESHOLD = 0.8
        self.opt = dataset_opt
        self.in_folder = dataset_opt['dataroot_LR']
        self.tar_folder = dataset_opt['dataroot_HR']
        self.mask_folder = dataset_opt['maskroot_HR']
        # self.ps = (32, 64, 64)
        with open(dataset_opt['uids_location']) as f:
            self.uids = [x.rstrip() for x in f.readlines()]
        self.scale = 1.
        self.ToTensor = util.ImgToTensor()

    def _cvt_int(self, inputString):
        find_digit = re.search(r'\d', inputString)
        reformat_num = int(float(inputString[find_digit.start():]))
        reformat_str = inputString[:find_digit.start()] + str(reformat_num)
        return reformat_str

    def __getitem__(self, index):
        uid = self.uids[index]
        uid_to_open_for_mask_target = uid.split('_')[0]

        # body mask - first we look at if the random voxel contain 80% of the body mask
        vol_mask = None
        if self.mask_folder:
            with h5py.File(os.path.join(self.mask_folder, uid_to_open_for_mask_target+'.h5'), 'r') as file:
                IMG_THICKNESS, IMG_WIDTH, IMG_HEIGHT = file['data'].shape
                vol_mask = file['data'][()]
        
        # LR
        with h5py.File(os.path.join(self.in_folder, uid+'.h5'), 'r') as file:
            vol_in = file['data'][()]
        
        # HR (note: HR cube indexes will be same as mask cube indexes since the masks are for HR images)
        vol_tar = None
        if self.tar_folder:
            with h5py.File(os.path.join(self.tar_folder, uid_to_open_for_mask_target+'.h5'), 'r') as file:
                vol_tar = file['data'][()]

        # creates an extra dimension. e.g. 32x64x64 becomes 1x32x64x64
        vol_in = np.expand_dims(vol_in, axis=0)
        vol_tar = np.expand_dims(vol_tar, axis=0)
        vol_mask = np.expand_dims(vol_mask, axis=0)

        # convert to tensors and also within a certain range of values as defined in 'self.ToTensor()' class
      
        vol_in, vol_tar = self.ToTensor(vol_in, clip=True, raw_data_range=1500.), self.ToTensor(vol_tar, clip=True, raw_data_range=1500.)
        vol_mask = self.ToTensor(vol_mask, clip=False, raw_data_range = 1)

        # read LR spacings
        spacings = [] 
        config_path = os.path.join(self.in_folder, uid + '.json')
        meta_data = util.read_config(config_path) # we only need spacing information for the configuration file
        spacings = meta_data['Spacing']

        out_dict = {'LR': vol_in, 'HR': vol_tar, 'mask': vol_mask, 'spacings': spacings, 'uid': uid}        

        return out_dict

    def __len__(self):
        return len(self.uids)
